#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import warnings
import matplotlib
if not hasattr(sys, "ps1"):
    matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy as np

# plt.style.use("seaborn-ticks")
plt.style.use("fast")
plt.rcParams["xtick.direction"] = "in"
plt.rcParams["ytick.direction"] = "in"
plt.rcParams["font.size"] = 11.0
plt.rcParams["figure.figsize"] = (9, 6)
plt.rcParams["legend.frameon"] = False


def line_plot(df, title=None, xlabel=None, ylabel="Cases",
              v=None, h=None,
              xlim=(None, None), ylim=(0, None),
              math_scale=True, x_logscale=False, y_logscale=False, y_integer=False,
              show_legend=True, bbox_to_anchor=(1.02, 0), bbox_loc="lower left",
              colormap=None, color_dict=None,
              filename=None):
    """
    Show chronological change of the data.

    Args:
        df (pandas.DataFrame): target data

            Index
                reset index
            Columns
                field names
            Values:
                data values
        title (str): title of the figure
        xlabel (str): x-label
        ylabel (str): y-label
        v (list[int/float]): list of x values of vertical lines or None
        h (list[int/float]): list of y values of horizontal lines or None
        xlim (tuple(int or float, int or float)): limit of x dimain
        ylim (tuple(int or float, int or float)): limit of y dimain
        math_scale (bool): whether use LaTEX or not
        x_logscale (bool): whether use log-scale in x-axis or not
        y_logscale (bool): whether use log-scale in y-axis or not
        y_integer (bool): whether force to show the values as integer or not
        show_legend (bool): whether show legend or not
        bbox_to_anchor (tuple(int or float, int or float)): distance of legend and plot
        bbox_loc (str): location of legend
        colormap (str, matplotlib colormap object or None): colormap, please refer to https://matplotlib.org/examples/color/colormaps_reference.html
        color_dict (dict[str, str] or None): dictionary of column names (keys) and colors (values)
        filename (str): filename of the figure, or None (show figure)

    Note:
        If None is included in xlim/ylim, the values will be automatically determined by Matplotlib
    """
    # Color
    if color_dict is None:
        color_args = {"colormap": colormap}
    else:
        colors = [color_dict.get(col) for col in df.columns]
        color_args = {"colormap": colormap, "color": colors}
    try:
        ax = df.plot(**color_args)
    except ValueError as e:
        raise ValueError(e.args[0]) from None
    # Scale
    if math_scale:
        ax.yaxis.set_major_formatter(
            matplotlib.ticker.ScalarFormatter(useMathText=True)
        )
        ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
    if x_logscale:
        ax.set_xscale("log")
        if xlim[0] == 0:
            xlim = (None, None)
    if y_logscale:
        ax.set_yscale("log")
        if ylim[0] == 0:
            ylim = (None, None)
    if y_integer:
        fmt = matplotlib.ticker.ScalarFormatter(useOffset=False)
        fmt.set_scientific(False)
        ax.yaxis.set_major_formatter(fmt)
    # Set metadata of figure
    ax.set_title(title or "")
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    if show_legend:
        ax.legend(bbox_to_anchor=bbox_to_anchor, loc=bbox_loc, borderaxespad=0)
    else:
        ax.legend().set_visible(False)
    if h is not None:
        ax.axhline(y=h, color="black", linestyle=":")
    if v is not None:
        if not isinstance(v, list):
            v = [v]
        for value in v:
            ax.axvline(x=value, color="black", linestyle=":")
    plt.tight_layout()
    # Save figure or show figure
    if filename is None:
        plt.show()
        return None
    plt.savefig(
        filename, bbox_inches="tight", transparent=False, dpi=300
    )
    plt.clf()
    return None


def box_plot(df, title, xlabel=None, ylabel=None,
             v=None, h=None,
             show_legend=True, bbox_to_anchor=(1.02, 0), bbox_loc="lower left",
             filename=None):
    """
    Show box plot of the data.

    Args:
        df (pandas.DataFrame): target data

            Index
                reset index
            Columns
                field names
            Values:
                data values
        title (str): title of the figure
        xlabel (str): x-label
        ylabel (str): y-label
        v (list[int/float]): list of x values of vertical lines or None
        h (list[int/float]): list of y values of horizontal lines or None
        show_legend (bool): whether show legend or not
        bbox_to_anchor (tuple(int or float, int or float)): distance of legend and plot
        bbox_loc (str): location of legend
        filename (str): filename of the figure, or None (show figure)
    """
    df.plot.bar(title=title)
    plt.xticks(rotation=0)
    if h is not None:
        plt.axhline(y=h, color="black", linestyle=":")
    plt.legend(
        bbox_to_anchor=bbox_to_anchor, loc=bbox_loc, borderaxespad=0
    )
    plt.tight_layout()
    if filename is None:
        plt.show()
        return None
    plt.savefig(
        filename, bbox_inches="tight", transparent=False, dpi=300
    )
    plt.clf()
    return None


def line_plot_multiple(df, x_col, actual_col, predicted_cols,
                       title, ylabel, xlim=(None, None), v=None, y_logscale=False, filename=None):
    """
    show multiple line graph of chronological change with actual plots.

    Args:
        df (pandas.DataFrame): data

            Index
                Date (pandas.TimeStamp): Observation date
            Columns
                - column defined by @x_col, values for x-axis
                - column defined by @actual_col, actual values for y-axis
                - columns defined by @predicted_cols, predicted values for y-axis
        x_col (str): column name for x-axis
        actual_col (str): column name for y-axis
        predicted_cols (list[str]): list of columns which have predicted values
        title (str): title of the figure
        y_label (str): label for y-axis
        xlim (tuple(int or float, int or float)): limit of x dimain
        y_logscale (bool): whether use log-scale in y-axis or not
        v (list[int]): list of Recovered values to show vertical lines
        filename (str): filename of the figure, or None (show figure)

    Note:
        When xlim[0] is None and lower x-axis limit determined by matplotlib automatically is lower than 0,
        lower x-axis limit will be set to 0.
    """
    warnings.simplefilter("ignore", FutureWarning)
    warnings.simplefilter("ignore", UserWarning)
    x_series = df[x_col]
    actual = df[actual_col]
    # Plot the actual values
    plt.plot(
        x_series, actual,
        label=actual_col, color="black", marker=".", markeredgewidth=0, linewidth=0)
    # Plot the predicted values
    for col in predicted_cols:
        plt.plot(x_series, df[col], label=col)
    # x-axis
    plt.xlabel(x_col)
    plt.xlim(max(plt.xlim()[0], xlim[0] or 0), xlim[1])
    # y-axis
    plt.ylabel(ylabel)
    try:
        plt.yscale("log", base=10)
    except Exception:
        plt.yscale("log", basey=10)
    # Delete y-labels of log-scale (minor) axis
    plt.setp(plt.gca().get_yticklabels(minor=True), visible=False)
    plt.gca().tick_params(left=False, which="minor")
    # Set new y-labels of major axis
    if y_logscale:
        ymin, ymax = plt.ylim()
        ydiff_scale = int(np.log10(ymax - ymin))
        yticks = np.linspace(
            round(ymin, - ydiff_scale),
            round(ymax, - ydiff_scale), 5, dtype=np.int64)
        plt.gca().set_yticks(yticks)
        fmt = matplotlib.ticker.ScalarFormatter(useOffset=False)
        fmt.set_scientific(False)
        plt.gca().yaxis.set_major_formatter(fmt)
    # Title
    plt.title(title)
    # Vertical lines
    if isinstance(v, (list, tuple)):
        for value in v:
            plt.axvline(x=value, color="black", linestyle=":")
    # Legend
    plt.legend(
        bbox_to_anchor=(1.02, 0), loc="lower left", borderaxespad=0
    )
    # Save figure or show figure
    plt.tight_layout()
    if filename is None:
        plt.show()
        return None
    plt.savefig(
        filename, bbox_inches="tight", transparent=False, dpi=300
    )
    plt.clf()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
from datetime import timedelta
from math import log10, floor
import warnings
import sys
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn import linear_model
from sklearn.exceptions import ConvergenceWarning
from sklearn.metrics import r2_score
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from covsirphy.util.error import deprecate, ScenarioNotFoundError, UnExecutedError
from covsirphy.util.error import NotRegisteredMainError, NotRegisteredExtraError
from covsirphy.util.plotting import line_plot, box_plot
from covsirphy.util.error import NotInteractiveError
from covsirphy.util.term import Term
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.analysis.param_tracker import ParamTracker
from covsirphy.analysis.data_handler import DataHandler


class Scenario(Term):
    """
    Scenario analysis.

    Args:
        jhu_data (covsirphy.JHUData or None): object of records
        population_data (covsirphy.PopulationData or None): PopulationData object
        country (str): country name (must not be None)
        province (str or None): province name
        tau (int or None): tau value
        auto_complement (bool): if True and necessary, the number of cases will be complemented

    Note:
        @jhu_data and @population_data must be registered with Scenario.register() if not specified here.
    """

    def __init__(self, jhu_data=None, population_data=None,
                 country=None, province=None, tau=None, auto_complement=True):
        # Area name
        if country is None:
            raise ValueError("@country must be specified.")
        self.country = str(country)
        self.province = str(province or self.UNKNOWN)
        self.area = JHUData.area_name(country, province)
        # Initialize data handler
        self._data = DataHandler(country=self.country, province=self.province)
        self._data.switch_complement(whether=auto_complement)
        # Tau value
        self.tau = self._ensure_tau(tau)
        # Register datasets
        self._tracker_dict = {}
        self.register(jhu_data=jhu_data, population_data=population_data, extras=None)
        # Initialize parameter tracker
        try:
            self._init_trackers()
        except NotRegisteredMainError:
            pass
        # Interactive (True) / script (False) mode
        self._interactive = hasattr(sys, "ps1")
        # Prediction of parameter values in the future phases: {name: (regression model, X_target)}
        self._lm_dict = {}

    def __getitem__(self, key):
        """
        Return the phase series object for the scenario.

        Args:
            key (str): scenario name

        Raises:
            ScenarioNotFoundError: the scenario is not registered

        Returns:
            covsirphy.PhaseSeries
        """
        if key in self._tracker_dict:
            return self._tracker_dict[key].series
        raise ScenarioNotFoundError(key)

    def __setitem__(self, key, value):
        """
        Register a phase series.

        Args:
            key (str): scenario name
            value (covsirphy.PhaseSeries): phase series object
        """
        self._tracker_dict[key] = ParamTracker(
            self._data.records(extras=False), value, area=self.area, tau=self.tau)

    @property
    def first_date(self):
        """
        str: the first date of the records
        """
        return self._data.first_date

    @first_date.setter
    def first_date(self, date):
        self._data.timepoints(
            first_date=date, last_date=self._data.last_date, today=self._data.today)

    @property
    def last_date(self):
        """
        str: the last date of the records
        """
        return self._data.last_date

    @last_date.setter
    def last_date(self, date):
        try:
            self._ensure_date_order(self._data.today, date)
        except ValueError:
            today = date
        else:
            today = self._data.today
        self.timepoints(
            first_date=self._data.first_date, last_date=date, today=today)

    @property
    def today(self):
        """
        str: reference date to determine whether a phase is a past phase or a future phase
        """
        return self._data.today

    @today.setter
    def today(self, date):
        self.timepoints(
            first_date=self._data.first_date, last_date=self._data.last_date, today=date)

    @property
    def interactive(self):
        """
        bool: interactive mode (display figures) or not

        Note:
            When running scripts, interactive mode cannot be selected.
        """
        return self._interactive

    @interactive.setter
    def interactive(self, is_interactive):
        if not hasattr(sys, "ps1") and is_interactive:
            raise NotInteractiveError
        self._interactive = hasattr(sys, "ps1") and bool(is_interactive)

    def register(self, jhu_data=None, population_data=None, extras=None):
        """
        Register datasets.

        Args:
            jhu_data (covsirphy.JHUData or None): object of records
            population_data (covsirphy.PopulationData or None): PopulationData object
            extras (list[covsirphy.CleaningBase] or None): extra datasets

        Raises:
            TypeError: non-data cleaning instance was included
            UnExpectedValueError: instance of un-expected data cleaning class was included as an extra dataset
        """
        self._data.register(jhu_data=jhu_data, population_data=population_data, extras=extras)
        if self._data.main_satisfied and not self._tracker_dict:
            self.timepoints()

    def timepoints(self, first_date=None, last_date=None, today=None):
        """
        Set the range of data and reference date to determine past/future of phases.

        Args:
            first_date (str or None): the first date of the records or None (min date of main dataset)
            last_date (str or None): the first date of the records or None (max date of main dataset)
            today (str or None): reference date to determine whether a phase is a past phase or a future phase

        Raises:
            NotRegisteredMainError: either JHUData or PopulationData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data

        Note:
            When @today is None, the reference date will be the same as @last_date (or max date).
        """
        self._data.timepoints(first_date=first_date, last_date=last_date, today=today)
        self._init_trackers()

    def line_plot(self, df, show_figure=True, filename=None, **kwargs):
        """
        Display or save a line plot of the dataframe.

        Args:
            show_figure (bool): whether show figure when interactive mode or not
            filename (str or None): filename of the figure or None (not save) when script mode

        Note:
            When interactive mode and @show_figure is True, display the figure.
            When script mode and filename is not None, save the figure.
            When using interactive shell, we can change the modes by Scenario.interactive = True/False.
        """
        if self._interactive and show_figure:
            return line_plot(df=df, filename=None, **kwargs)
        if not self._interactive and filename is not None:
            return line_plot(df=df, filename=filename, **kwargs)

    def complement(self, **kwargs):
        """
        Complement the number of recovered cases, if necessary.

        Args:
            kwargs: the other arguments of JHUData.subset_complement()

        Returns:
            covsirphy.Scenario: self
        """
        self._data.switch_complement(whether=True, **kwargs)
        return self

    def complement_reverse(self):
        """
        Restore the raw records. Reverse method of covsirphy.Scenario.complement().

        Returns:
            covsirphy.Scenario: self
        """
        self._data.switch_complement(whether=False)
        return self

    def show_complement(self, **kwargs):
        """
        Show the details of complement that was (or will be) performed for the records.

        Args:
            kwargs: keyword arguments of JHUDataComplementHandler() i.e. control factors of complement

        Returns:
            pandas.DataFrame: as the same as JHUData.show_complement()
        """
        self._data.switch_complement(whether=None, **kwargs)
        return self._data.show_complement()

    def records(self, variables=None, **kwargs):
        """
        Return the records as a dataframe.

        Args:
            variables (list[str] or str or None): list of variables, 'all', None (Infected/Fatal/Recovered)
            kwargs: the other keyword arguments of Scenario.line_plot()

        Raises:
            NotRegisteredMainError: either JHUData or PopulationData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            NotRegisteredExtraError: some variables are not included in the main datasets
            and no extra datasets were registered

        Returns:
            pandas.DataFrame

                Index
                    reset index
                Columns
                    - Date (pd.TimeStamp): Observation date
                    - Columns set by @variables (int)

        Note:
            - Records with Recovered > 0 will be selected.
            - If complement was performed by Scenario.complement() or Scenario(auto_complement=True),
            The kind of complement will be added to the title of the figure.
            - @variables can be selected from Susceptible/Confirmed/Infected/Fatal/Recovered.
        """
        # Get necessary data for the variables
        all_df = self._data.records_all()
        if variables is None:
            df = all_df.loc[:, [self.DATE, self.CI, self.F, self.R]]
        elif variables == "all":
            df = all_df.copy()
        else:
            self._ensure_list(variables, candidates=all_df.columns.tolist(), name="variables")
            df = all_df.loc[:, [self.DATE, *variables]]
        # Figure
        if self._data.complemented:
            title = f"{self.area}: Cases over time\nwith {self._data.complemented}"
        else:
            title = f"{self.area}: Cases over time"
        self.line_plot(
            df=df.set_index(self.DATE), title=title, y_integer=True, **kwargs)
        return df

    def records_diff(self, variables=None, window=7, **kwargs):
        """
        Return the number of daily new cases (the first discreate difference of records).

        Args:
            variables (list[str] or str or None): list of variables, 'all', None (Infected/Fatal/Recovered)
            window (int): window of moving average, >= 1
            kwargs: the other keyword arguments of Scenario.line_plot()

        Returns:
            pandas.DataFrame
                Index
                    - Date (pd.TimeStamp): Observation date
                Columns
                    - Confirmed (int): daily new cases of Confirmed, if calculated
                    - Infected (int):  daily new cases of Infected, if calculated
                    - Fatal (int):  daily new cases of Fatal, if calculated
                    - Recovered (int):  daily new cases of Recovered, if calculated

        Note:
            @variables will be selected from Confirmed, Infected, Fatal and Recovered.
            If None was set as @variables, ["Confirmed", "Fatal", "Recovered"] will be used.
        """
        window = self._ensure_natural_int(window, name="window")
        df = self.records(variables=variables, show_figure=False).set_index(self.DATE)
        df = df.diff().dropna()
        df = df.rolling(window=window).mean().dropna().astype(np.int64)
        if self._data.complemented:
            title = f"{self.area}: Daily new cases\nwith {self._data.complemented}"
        else:
            title = f"{self.area}: Daily new cases"
        self.line_plot(df=df, title=title, y_integer=True, **kwargs)
        return df

    def _init_trackers(self):
        """
        Initialize dictionary of trackers.
        """
        data = copy.deepcopy(self._data)
        series = ParamTracker.create_series(
            first_date=data.first_date, last_date=data.last_date, population=data.population)
        tracker = ParamTracker(
            record_df=self._data.records(extras=False), phase_series=series, area=self.area, tau=self.tau)
        self._tracker_dict = {self.MAIN: tracker}

    def _tracker(self, name, template="Main"):
        """
        Ensure that the phases series is registered.
        If not registered, copy the template phase series.

        Args:
            name (str): phase series name
            template (str): name of template phase series

        Returns:
            covsirphy.ParamTracker
        """
        # Registered
        if name in self._tracker_dict:
            return self._tracker_dict[name]
        # Un-registered and create it
        if template not in self._tracker_dict:
            raise ScenarioNotFoundError(template)
        tracker = copy.deepcopy(self._tracker_dict[template])
        self._tracker_dict[name] = tracker
        return tracker

    @deprecate(old="Scenario.add_phase()", new="Scenario.add()")
    def add_phase(self, **kwargs):
        return self.add(**kwargs)

    def add(self, name="Main", end_date=None, days=None, population=None, model=None, **kwargs):
        """
        Add a new phase.
        The start date will be the next date of the last registered phase.

        Args:
            name (str): phase series name, 'Main' or user-defined name
            end_date (str): end date of the new phase
            days (int): the number of days to add
            population (int or None): population value of the start date
            model (covsirphy.ModelBase or None): ODE model
            kwargs: keyword arguments of ODE model parameters, not including tau value.

        Returns:
            covsirphy.Scenario: self

        Note:
            - If the phases series has not been registered, new phase series will be created.
            - Either @end_date or @days must be specified.
            - If @end_date and @days are None, the end date will be the last date of the records.
            - If both of @end_date and @days were specified, @end_date will be used.
            - If @popultion is None, initial value will be used.
            - If @model is None, the model of the last phase will be used.
            - Tau will be fixed as the last phase's value.
            - kwargs: Default values are the parameter values of the last phase.
        """
        if end_date is not None:
            self._ensure_date(end_date, name="end_date")
        tracker = self._tracker(name)
        try:
            tracker.add(
                end_date=end_date, days=days, population=population, model=model, **kwargs)
        except ValueError:
            last_date = tracker.series.unit("last").end_date
            raise ValueError(
                f'@end_date must be over {last_date}. However, {end_date} was applied.') from None
        self._tracker_dict[name] = tracker
        return self

    def clear(self, name="Main", include_past=False, template="Main"):
        """
        Clear phase information.

        Args:
            name (str): scenario name
            include_past (bool): if True, past phases will be removed as well as future phases
            template (str): name of template scenario

        Returns:
            covsirphy.Scenario: self

        Note:
            If un-registered scenario name was specified, new scenario will be created.
            Future phases will be always deleted.
        """
        tracker = self._tracker(name, template=template)
        if include_past:
            self[name] = tracker.delete_all()
        else:
            self[name] = tracker.delete(phases=tracker.future_phases()[0])
        return self

    def _delete_series(self, name):
        """
        Delete a scenario or initialise main scenario.

        Args:
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        if name == self.MAIN:
            self[self.MAIN] = self._tracker(self.MAIN).delete_all()
        else:
            self._tracker_dict.pop(name)
        return self

    def delete(self, phases=None, name="Main"):
        """
        Delete phases.

        Args:
            phase (list[str] or None): phase names, or ['last']
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self

        Note:
            If @phases is None, the phase series will be deleted.
            When @phase is '0th', disable 0th phase. 0th phase will not be deleted.
            If the last phase is included in @phases, the dates will be released from phases.
            If the last phase is not included, the dates will be assigned to the previous phase.
        """
        # Clear main series or delete sub phase series
        if phases is None:
            return self._delete_series(name)
        # Delete phases
        tracker = self._tracker(name)
        self[name] = tracker.delete(phases=phases)
        return self

    def disable(self, phases, name="Main"):
        """
        The phases will be disabled and removed from summary.

        Args:
            phase (list[str] or None): phase names or None (all enabled phases)
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).disable(phases)
        return self

    def enable(self, phases, name="Main"):
        """
        The phases will be enabled and appear in summary.

        Args:
            phase (list[str] or None): phase names or None (all disabled phases)
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).enable(phases)
        return self

    def combine(self, phases, name="Main", population=None, **kwargs):
        """
        Combine the sequential phases as one phase.
        New phase name will be automatically determined.

        Args:
            phases (list[str]): list of phases
            name (str, optional): name of phase series
            population (int): population value of the start date
            kwargs: keyword arguments to save as phase information

        Raises:
            TypeError: @phases is not a list

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).combine(
            phases=phases, population=population, **kwargs)
        return self

    def separate(self, date, name="Main", population=None, **kwargs):
        """
        Create a new phase with the change point.
        New phase name will be automatically determined.

        Args:
            date (str): change point, i.e. start date of the new phase
            name (str): scenario name
            population (int): population value of the change point
            kwargs: keyword arguments of PhaseUnit.set_ode() if update is necessary

        Returns:
            covsirphy.Scenario: self
        """
        self[name] = self._tracker(name).separate(
            date=date, population=population, **kwargs)
        return self

    def _summary(self, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown

        Returns:
            pandas.DataFrame:
            - if @name not None, as the same as PhaseSeries().summary()
            - if @name is None, index will be phase series name and phase name

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        if name is None:
            if len(self._tracker_dict.keys()) > 1:
                dataframes = []
                for (_name, tracker) in self._tracker_dict.items():
                    summary_df = tracker.series.summary()
                    summary_df = summary_df.rename_axis(self.PHASE)
                    summary_df[self.SERIES] = _name
                    dataframes.append(summary_df.reset_index())
                df = pd.concat(dataframes, ignore_index=True, sort=False)
                return df.set_index([self.SERIES, self.PHASE])
            name = self.MAIN
        return self._tracker(name).series.summary()

    def summary(self, columns=None, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown
            columns (list[str] or None): columns to show

        Returns:
            pandas.DataFrame:
            - if @name not None, as the same as PhaseSeries().summary()
            - if @name is None, index will be phase series name and phase name

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
            If @columns is None, all columns will be shown.
        """
        df = self._summary(name=name)
        all_cols = df.columns.tolist()
        if set(self.EST_COLS).issubset(all_cols):
            all_cols = [col for col in all_cols if col not in self.EST_COLS]
            all_cols += self.EST_COLS
        columns = columns or all_cols
        self._ensure_list(columns, candidates=all_cols, name="columns")
        df = df.loc[:, columns]
        return df.dropna(how="all", axis=1).fillna(self.UNKNOWN)

    def trend(self, force=True, name="Main", show_figure=True, filename=None, **kwargs):
        """
        Perform S-R trend analysis and set phases.

        Args:
            force (bool): if True, change points will be over-written
            name (str): phase series name
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (display)
            kwargs: keyword arguments of covsirphy.ChangeFinder() and covsirphy.line_plot_multiple()

        Returns:
            covsirphy.Scenario: self
        """
        # Arguments
        if "n_points" in kwargs.keys():
            raise ValueError(
                "@n_points argument is un-necessary"
                " because the number of change points will be automatically determined.")
        try:
            include_init_phase = kwargs.pop("include_init_phase")
            warnings.warn(
                "@include_init_phase was deprecated. Please use Scenario.disable('0th').",
                DeprecationWarning, stacklevel=2)
        except KeyError:
            include_init_phase = True
        try:
            force = kwargs.pop("set_phases")
        except KeyError:
            pass
        # S-R trend analysis
        tracker = self._tracker(name)
        if not self._interactive and filename is None:
            show_figure = False
        filename = None if self._interactive else filename
        self[name] = tracker.trend(
            force=force, show_figure=show_figure, filename=filename, **kwargs)
        # Disable 0th phase, if necessary
        if not include_init_phase:
            self[name] = tracker.disable(phases=["0th"])
        return self

    def estimate(self, model, phases=None, name="Main", n_jobs=-1, **kwargs):
        """
        Perform parameter estimation for each phases.

        Args:
            model (covsirphy.ModelBase): ODE model
            phases (list[str]): list of phase names, like 1st, 2nd...
            name (str): phase series name
            n_jobs (int): the number of parallel jobs or -1 (CPU count)
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()

        Note:
            - If 'Main' was used as @name, main PhaseSeries will be used.
            - If @name phase was not registered, new PhaseSeries will be created.
            - If @phases is None, all past phase will be used.
            - Phases with estimated parameter values will be ignored.
            - In kwargs, tau value cannot be included.
        """
        if self.TAU in kwargs:
            raise ValueError(
                "@tau must be specified when scenario = Scenario(), and cannot be specified here.")
        self.tau, self[name] = self._tracker(name).estimate(
            model=model, phases=phases, n_jobs=n_jobs, **kwargs)

    def phase_estimator(self, phase, name="Main"):
        """
        Return the estimator of the phase.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name

        Return:
            covsirphy.Estimator: estimator of the phase
        """
        estimator = self._tracker_dict[name].series.unit(phase).estimator
        if estimator is None:
            raise UnExecutedError(f'Scenario.estimate(model, phases=["{phase}"], name={name})')
        return estimator

    def estimate_history(self, phase, name="Main", **kwargs):
        """
        Show the history of optimization.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.Estimator.history()

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        estimator = self.phase_estimator(phase=phase, name=name)
        estimator.history(**kwargs)

    def estimate_accuracy(self, phase, name="Main", **kwargs):
        """
        Show the accuracy as a figure.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.Estimator.accuracy()

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        estimator = self.phase_estimator(phase=phase, name=name)
        estimator.accuracy(**kwargs)

    def simulate(self, variables=None, phases=None, name="Main", y0_dict=None, **kwargs):
        """
        Simulate ODE models with set/estimated parameter values and show it as a figure.

        Args:
            variables (list[str] or None): variables to include, Infected/Fatal/Recovered when None
            phases (list[str] or None): phases to shoe or None (all phases)
            name (str): phase series name. If 'Main', main PhaseSeries will be used
            y0_dict(dict[str, float] or None): dictionary of initial values of variables
            kwargs: the other keyword arguments of Scenario.line_plot()

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pd.TimeStamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - Variables of the model and dataset (int): Confirmed etc.
        """
        tracker = copy.deepcopy(self._tracker(name))
        # Select phases
        if phases is not None:
            tracker.disable(phases=None)
            tracker.enable(phases=phases)
        # Simulation
        try:
            sim_df = tracker.simulate(y0_dict=y0_dict)
        except UnExecutedError:
            raise UnExecutedError(
                "Scenario.trend() or Scenario.add(), and Scenario.estimate(model)") from None
        # Show figure
        df = sim_df.set_index(self.DATE)
        fig_cols = self._ensure_list(
            variables or [self.CI, self.F, self.R], candidates=df.columns.tolist(), name="variables")
        title = f"{self.area}: Simulated number of cases ({name} scenario)"
        self.line_plot(df=df[fig_cols], title=title, y_integer=True, v=tracker.change_dates(), **kwargs)
        return sim_df

    def get(self, param, phase="last", name="Main"):
        """
        Get the parameter value of the phase.

        Args:
            param (str): parameter name (columns in self.summary())
            phase (str): phase name or 'last'
                - if 'last', the value of the last phase will be returned
            name (str): phase series name

        Returns:
            str or int or float

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        df = self.summary(name=name)
        if param not in df.columns:
            raise KeyError(f"@param must be in {', '.join(df.columns)}.")
        if phase == "last":
            phase = df.index[-1]
        return df.loc[phase, param]

    def _param_history(self, targets, name):
        """
        Return the subset of summary dataframe to select the target of parameter history.

        Args:
            targets (list[str] or str): parameters to show (Rt etc.)
            name (str): phase series name

        Returns:
            pandas.DataFrame: selected summary dataframe

        Raises:
            KeyError: targets are not in the columns of summary dataframe
        """
        series = self._tracker_dict[name].series
        model_set = {unit.model for unit in series}
        model_set = model_set - set([None])
        parameters = self.flatten([m.PARAMETERS for m in model_set])
        day_params = self.flatten([m.DAY_PARAMETERS for m in model_set])
        selectable_cols = [self.N, *parameters, self.RT, *day_params]
        selectable_set = set(selectable_cols)
        df = series.summary().replace(self.UNKNOWN, None)
        if not selectable_set.issubset(df.columns):
            raise UnExecutedError(
                f'Scenario.estimate(model, phases=None, name="{name}")')
        targets = [targets] if isinstance(targets, str) else targets
        targets = targets or selectable_cols
        if not set(targets).issubset(selectable_set):
            raise KeyError(
                f"@targets must be selected from {', '.join(selectable_cols)}."
            )
        df = df.loc[:, targets].dropna(how="any", axis=0)
        return df.astype(np.float64)

    @deprecate(
        old="Scenario.param_history(targets: list)",
        new="Scenario.history(target: str)",
        version="2.7.3-alpha")
    def param_history(self, targets=None, name="Main", divide_by_first=True,
                      show_figure=True, filename=None, show_box_plot=True, **kwargs):
        """
        Return subset of summary and show a figure to show the history.

        Args:
            targets (list[str] or str): parameters to show (Rt etc.)
            name (str): phase series name
            divide_by_first (bool): if True, divide the values by 1st phase's values
            show_box_plot (bool): if True, box plot. if False, line plot
            show_figure (bool): If True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)
            kwargs: keyword arguments of pd.DataFrame.plot or line_plot()

        Returns:
            pandas.DataFrame

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        self._tracker(name)
        # Select target to show
        df = self._param_history(targets, name)
        # Divide by the first phase parameters
        if divide_by_first:
            df = df / df.iloc[0, :]
            title = f"{self.area}: Ratio to 1st phase parameters ({name} scenario)"
        else:
            title = f"{self.area}: History of parameter values ({name} scenario)"
        if not show_figure:
            return df
        if show_box_plot:
            h_values = [1.0] if divide_by_first or self.RT in targets else None
            box_plot(df, title, h=h_values, filename=filename)
            return df
        _df = df.reset_index(drop=True)
        _df.index = _df.index + 1
        h = 1.0 if divide_by_first else None
        line_plot(
            _df, title=title,
            xlabel="Phase", ylabel=str(), math_scale=False, h=h,
            filename=filename
        )
        return df

    def _describe(self, y0_dict=None):
        """
        Describe representative values.

        Args:
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame
                Index
                    (int): scenario name
                Columns
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Confirmed({date}): Confirmed on the next date of the last phase
                    - Infected({date}): Infected on the next date of the last phase
                    - Fatal({date}): Fatal on the next date of the last phase
        """
        _dict = {}
        for (name, _) in self._tracker_dict.items():
            # Predict the number of cases
            df = self.simulate(name=name, y0_dict=y0_dict, show_figure=False)
            df = df.set_index(self.DATE)
            cols = df.columns[:]
            last_date = df.index[-1]
            # Max value of Infected
            max_ci = df[self.CI].max()
            argmax_ci = df[self.CI].idxmax().strftime(self.DATE_FORMAT)
            # Confirmed on the next date of the last phase
            last_c = df.loc[last_date, self.C]
            # Infected on the next date of the last phase
            last_ci = df.loc[last_date, self.CI]
            # Fatal on the next date of the last phase
            last_f = df.loc[last_date, self.F] if self.F in cols else None
            # Save representative values
            last_date_str = last_date.strftime(self.DATE_FORMAT)
            _dict[name] = {
                f"max({self.CI})": max_ci,
                f"argmax({self.CI})": argmax_ci,
                f"{self.C} on {last_date_str}": last_c,
                f"{self.CI} on {last_date_str}": last_ci,
                f"{self.F} on {last_date_str}": last_f,
            }
        return pd.DataFrame.from_dict(_dict, orient="index")

    def describe(self, y0_dict=None, with_rt=True):
        """
        Describe representative values.

        Args:
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value
            with_rt (bool): whether show the history of Rt values

        Returns:
            pandas.DataFrame:
                Index
                    str: scenario name
                Columns
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Confirmed({date}): Confirmed on the next date of the last phase
                    - Infected({date}): Infected on the next date of the last phase
                    - Fatal({date}): Fatal on the next date of the last phase
                    - nth_Rt etc.: Rt value if the values are not the same values
        """
        df = self._describe(y0_dict=y0_dict)
        if not with_rt or len(self._tracker_dict) == 1:
            return df
        # History of reproduction number
        rt_df = self.summary().reset_index()
        rt_df = rt_df.pivot_table(index=self.SERIES, columns=self.PHASE, values=self.RT)
        rt_df = rt_df.fillna(self.UNKNOWN)
        rt_df = rt_df.loc[:, rt_df.nunique() > 1]
        cols = sorted(rt_df, key=self.str2num)
        return df.join(rt_df[cols].add_suffix(f"_{self.RT}"), how="left")

    def _track_param(self, name):
        """
        Get the history of parameters for the scenario.

        Args:
            name (str): phase series name

        Returns:
            pandas.DataFrame:
                Index Date (pandas.TimeStamp)
                Columns
                    - Population (int)
                    - Rt (float)
                    - parameter values (float)
                    - day parameter values (float)
        """
        df = self.summary(name=name).replace(self.UNKNOWN, None)
        # Date range to dates
        df[self.START] = pd.to_datetime(df[self.START])
        df[self.END] = pd.to_datetime(df[self.END])
        df[self.DATE] = df[[self.START, self.END]].apply(
            lambda x: pd.date_range(x[0], x[1]).tolist(), axis=1)
        df = df.reset_index(drop=True).explode(self.DATE)
        # Columns
        df = df.drop(
            [self.TENSE, self.START, self.END, self.ODE, self.TAU, *self.EST_COLS],
            axis=1, errors="ignore")
        df = df.set_index(self.DATE)
        for col in df.columns:
            df[col] = pd.to_numeric(df[col], errors="coerce")
        df[self.N] = df[self.N].astype(np.int64)
        return df

    def _track(self, phases=None, name="Main", y0_dict=None):
        """
        Show values of parameters and variables in one dataframe for the scenario.

        Args:
            phases (list[str] or None): phases to shoe or None (all phases)
            name (str): phase series name
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame:
                Index reset index
                Columns
                    - Date (pandas.TimeStamp)
                    - variables (int)
                    - Population (int)
                    - Rt (float)
                    - parameter values (float)
                    - day parameter values (float)
        """
        sim_df = self.simulate(phases=phases, name=name, y0_dict=y0_dict, show_figure=False)
        param_df = self._track_param(name=name)
        return pd.merge(
            sim_df, param_df, how="inner", left_on=self.DATE, right_index=True, sort=True)

    def track(self, phases=None, with_actual=True, y0_dict=None):
        """
        Show values of parameters and variables in one dataframe.

        Args:
            phases (list[str] or None): phases to shoe or None (all phases)
            with_actual (bool): if True, show actual number of cases will included as "Actual" scenario
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame: tracking records
                Index
                    reset index
                Columns
                    - Scenario (str)
                    - Date (pandas.TimeStamp)
                    - variables (int)
                    - Population (int)
                    - Rt (float)
                    - parameter values (float)
                    - day parameter values (float)
        """
        dataframes = []
        append = dataframes.append
        for name in self._tracker_dict.keys():
            df = self._track(phases=phases, name=name, y0_dict=y0_dict)
            df.insert(0, self.SERIES, name)
            append(df)
        if with_actual:
            df = self._data.records(extras=False)
            df.insert(0, self.SERIES, self.ACTUAL)
            append(df)
        return pd.concat(dataframes, axis=0, sort=False)

    def _history(self, target, phases=None, with_actual=True, y0_dict=None):
        """
        Show the history of variables and parameter values to compare scenarios.

        Args:
            target (str): parameter or variable name to show (Rt, Infected etc.)
            phases (list[str] or None): phases to shoe or None (all phases)
            with_actual (bool): if True and @target is a variable name, show actual number of cases
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value

        Returns:
            pandas.DataFrame
        """
        # Include actual data or not
        with_actual = with_actual and target in self.VALUE_COLUMNS
        # Get tracking data
        df = self.track(phases=phases, with_actual=with_actual, y0_dict=y0_dict)
        if target not in df.columns:
            col_str = ", ".join(list(df.columns))
            raise KeyError(f"@target must be selected from {col_str}, but {target} was applied.")
        # Select the records of target variable
        return df.pivot_table(
            values=target, index=self.DATE, columns=self.SERIES, aggfunc="last")

    def history(self, target, phases=None, with_actual=True, y0_dict=None, **kwargs):
        """
        Show the history of variables and parameter values to compare scenarios.

        Args:
            target (str): parameter or variable name to show (Rt, Infected etc.)
            phases (list[str] or None): phases to shoe or None (all phases)
            with_actual (bool): if True and @target is a variable name, show actual number of cases
            y0_dict (dict or None): dictionary of initial values or None
                - key (str): variable name
                - value (float): initial value
            kwargs: the other keyword arguments of Scenario.line_plot()

        Returns:
            pandas.DataFrame
        """
        df = self._history(target=target, phases=phases, with_actual=with_actual, y0_dict=y0_dict)
        df.dropna(subset=[col for col in df.columns if col != self.ACTUAL], inplace=True)
        if target == self.RT:
            ylabel = self.RT_FULL
        elif target in self.VALUE_COLUMNS:
            ylabel = f"The number of {target.lower()} cases"
        else:
            ylabel = target
        title = f"{self.area}: {ylabel} over time"
        tracker = self._tracker(self.MAIN)
        self.line_plot(
            df=df, title=title, ylabel=ylabel, v=tracker.change_dates(), math_scale=False,
            h=1.0 if target == self.RT else None, **kwargs)
        return df

    def history_rate(self, params=None, name="Main", **kwargs):
        """
        Show change rates of parameter values in one figure.
        We can find the parameters which increased/decreased significantly.

        Args:
            params (list[str] or None): parameters to show
            name (str): phase series name
            kwargs: the other keyword arguments of Scenario.line_plot()

        Returns:
            pandas.DataFrame
        """
        df = self._track_param(name=name)
        model = self._tracker(name).last_model
        cols = list(set(df.columns) & set(model.PARAMETERS))
        if params is not None:
            if not isinstance(params, (list, set)):
                raise TypeError(f"@params must be a list of parameters, but {params} were applied.")
            cols = list(set(cols) & set(params)) or cols
        df = df.loc[:, cols] / df.loc[df.index[0], cols]
        # Show figure
        f_date = df.index[0].strftime(self.DATE_FORMAT)
        title = f"{self.area}: {model.NAME} parameter change rates over time (1.0 on {f_date})"
        ylabel = f"Value per that on {f_date}"
        title = f"{self.area}: {ylabel} over time"
        tracker = self._tracker(self.MAIN)
        self.line_plot(
            df=df, title=title, ylabel=ylabel, v=tracker.change_dates(), math_scale=False, **kwargs)
        return df

    def retrospective(self, beginning_date, model, control="Main", target="Target", **kwargs):
        """
        Perform retrospective analysis.
        Compare the actual series of phases (control) and
        series of phases with specified parameters (target).

        Args:
            beginning_date (str): when the parameter values start to be changed from actual values
            model (covsirphy.ModelBase): ODE model
            control (str): scenario name of control
            target (str): scenario name of target
            kwargs: keyword argument of parameter values and Estimator.run()

        Note:
            When parameter values are not specified,
            actual values of the last date before the beginning date will be used.
        """
        param_dict = {k: v for (k, v) in kwargs.items() if k in model.PARAMETERS}
        est_kwargs = dict(kwargs.items() - param_dict.items())
        # Control
        self.clear(name=control, include_past=True)
        self.trend(name=control, show_figure=False)
        try:
            self.separate(date=beginning_date, name=control)
        except ValueError:
            pass
        self.estimate(model, name=control, **est_kwargs)
        # Target
        self.clear(name=target, include_past=False, template=control)
        phases_changed = [
            self.num2str(i) for (i, ph) in enumerate(self._tracker(target).series)
            if ph >= beginning_date]
        self.delete(phases=phases_changed, name=target)
        self.add(name=target, **param_dict)
        self.estimate(model, name=target, **est_kwargs)

    def score(self, metrics="RMSLE", variables=None, phases=None, past_days=None, name="Main", y0_dict=None):
        """
        Evaluate accuracy of phase setting and parameter estimation of all enabled phases all some past days.

        Args:
            metrics (str): "MAE", "MSE", "MSLE", "RMSE" or "RMSLE"
            variables (list[str] or None): variables to use in calculation
            phases (list[str] or None): phases to use in calculation
            past_days (int or None): how many past days to use in calculation, natural integer
            name(str): phase series name. If 'Main', main PhaseSeries will be used
            y0_dict(dict[str, float] or None): dictionary of initial values of variables

        Returns:
            float: score with the specified metrics

        Note:
            If @variables is None, ["Infected", "Fatal", "Recovered"] will be used.
            "Confirmed", "Infected", "Fatal" and "Recovered" can be used in @variables.
            If @phases is None, all phases will be used.
            @phases and @past_days can not be specified at the same time.
        """
        tracker = self._tracker(name)
        if past_days is not None:
            if phases is not None:
                raise ValueError(
                    "@phases and @past_days cannot be specified at the same time.")
            past_days = self._ensure_natural_int(past_days, name="past_days")
            # Separate a phase, if possible
            beginning_date = self.date_change(self._data.last_date, days=0 - past_days)
            try:
                tracker.separate(date=beginning_date)
            except ValueError:
                pass
            # Ge the list of target phases
            phases = [
                self.num2str(num) for (num, unit)
                in enumerate(tracker.series)
                if unit >= beginning_date
            ]
        return tracker.score(
            metrics=metrics, variables=variables, phases=phases, y0_dict=y0_dict)

    def estimate_delay(self, oxcgrt_data=None, indicator="Stringency_index",
                       target="Confirmed", value_range=(7, None)):
        """
        Estimate the average day [days] between the indicator and the target.
        We assume that the indicator impact on the target value with delay.

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset
            indicator (str): indicator name, a column of any registered datasets
            target (str): target name, a column of any registered datasets
            value_range (tuple(int, int or None)): tuple, giving the minimum and maximum range to search for change over time

        Raises:
            NotRegisteredMainError: either JHUData or PopulationData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            UserWarning: failed in calculating and returned the default value (recovery period)

        Returns:
            tuple(int, pandas.DataFrame):
                - int: the estimated number of days of delay [day]
                - pandas.DataFrame:
                    Index
                        reset index
                    Columns
                        - (int or float): column defined by @indicator
                        - (int or float): column defined by @target
                        - (int): column defined by @delay_name [days]

        Note:
            - Average recovered period of JHU dataset will be used as returned value when the estimated value was not in value_range.
            - Very long periods (outside of 99% quantile) are taken out.
            - @oxcgrt_data argument was deprecated. Please use Scenario.register(extras=[oxcgrt_data]).
        """
        # Register OxCGRT data
        if oxcgrt_data is not None:
            warnings.warn(
                "Please use Scenario.register(extras=[oxcgrt_data]) rather than Scenario.fit(oxcgrt_data).",
                DeprecationWarning, stacklevel=1)
            self.register(extras=[oxcgrt_data])
        # Calculate delay values
        df = self._data.estimate_delay(indicator=indicator, target=target, delay_name="Period Length")
        # Filter out very long periods
        df_filtered = df.loc[df["Period Length"] < df["Period Length"].quantile(0.99)]
        if value_range[1] is not None:
            df_filtered = df_filtered.loc[df["Period Length"] < value_range[1]]
        delay_days = self._data.recovery_period() if df_filtered.empty else int(df_filtered["Period Length"].mean())
        return (delay_days, df)

    def _fit_create_data(self, model, name, delay, removed_cols):
        """
        Create train/test dataset for Elastic Net regression,
        assuming that extra variables will impact on ODE parameter values with delay.

        Args:
            model (covsirphy.ModelBase): ODE model
            name (str): scenario name
            delay (int): delay period
            removed_cols (list): list of variables to remove from X dataset

        Returns:
            tuple(pandas.DataFrame):
                - X dataset for linear regression
                - y dataset for linear regression
                - X dataset of the target dates
        """
        # Clear the future phases
        self.clear(name=name, include_past=False)
        # Parameter values
        param_df = self._track_param(name=name)[model.PARAMETERS]
        # Extra datasets (explanatory variables)
        extras_df = self._data.records(main=False, extras=True).set_index(self.DATE)
        extras_df = extras_df.loc[:, ~extras_df.columns.isin(removed_cols)]
        # Apply delay on OxCGRT data
        extras_df.index += timedelta(days=delay)
        # Create training/test dataset
        df = param_df.join(extras_df, how="inner")
        df = df.rolling(window=delay).mean().dropna().drop_duplicates()
        X = df.drop(model.PARAMETERS, axis=1)
        y = df.loc[:, model.PARAMETERS]
        # X dataset of the target dates
        dates = pd.date_range(
            start=param_df.index.max() + timedelta(days=1),
            end=extras_df.index.max(),
            freq="D")
        X_target = extras_df.loc[dates]
        return (X, y, X_target)

    def fit(self, oxcgrt_data=None, name="Main", test_size=0.2, seed=0, delay=None):
        """
        Learn the relationship of ODE parameter values and delayed OxCGRT scores using Elastic Net regression,
        assuming that OxCGRT scores will impact on ODE parameter values with delay.
        Min-max scaling and Elastic net regression with parameter optimization and cross validation.

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset
            name (str): scenario name
            test_size (float): proportion of the test dataset of Elastic Net regression
            seed (int): random seed when spliting the dataset to train/test data
            delay (int): number of days of delay between policy measure and effect
            on number of confirmed cases.

        Raises:
            covsirphy.UnExecutedError: Scenario.estimate() or Scenario.add() were not performed

        Returns:
            dict(str, object):
                - scaler (object): scaler class
                - regressor (object): regressor class
                - alpha (float): alpha value used in Elastic Net regression
                - l1_ratio (float): l1_ratio value used in Elastic Net regression
                - score_train (float): determination coefficient of train dataset
                - score_test (float): determination coefficient of test dataset
                - X_train (numpy.array): X_train
                - y_train (numpy.array): y_train
                - X_test (numpy.array): X_test
                - y_test (numpy.array): y_test
                - X_target (numpy.array): X_target
                - intercept (pandas.DataFrame): intercept values (Index ODE parameters, Columns indicators)
                - delay (int): number of days of delay between policy measure and effect
                  on number of confirmed cases.

        Note:
            @oxcgrt_data argument was deprecated. Please use Scenario.register(extras=[oxcgrt_data]).
        """
        warnings.simplefilter("ignore", category=ConvergenceWarning)
        # Register OxCGRT data
        if oxcgrt_data is not None:
            warnings.warn(
                "Please use Scenario.register(extras=[oxcgrt_data]) rather than Scenario.fit(oxcgrt_data).",
                DeprecationWarning, stacklevel=1)
            self.register(extras=[oxcgrt_data])
        # ODE model
        model = self._tracker(name).last_model
        if model is None:
            raise UnExecutedError(
                "Scenario.estimate() or Scenario.add()",
                message=f", specifying @model (covsirphy.SIRF etc.) and @name='{name}'.")
        # Set delay effect
        if delay is None:
            delay, delay_df = self.estimate_delay(oxcgrt_data)
            removed_cols = delay_df.columns.tolist()
        else:
            delay = self._ensure_natural_int(delay, name="delay")
            removed_cols = []
        # Create training/test dataset
        try:
            X, y, X_target = self._fit_create_data(
                model=model, name=name, delay=delay, removed_cols=removed_cols)
        except NotRegisteredExtraError:
            raise NotRegisteredExtraError(
                "Scenario.register(jhu_data, population_data, extras=[...])",
                message="with extra datasets") from None
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=seed)
        # Create pipeline for learning
        cv = linear_model.MultiTaskElasticNetCV(
            alphas=[0, 0.001, 0.01, 0.1, 1, 10, 100, 1000],
            l1_ratio=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            cv=5, n_jobs=-1)
        steps = [
            ("scaler", MinMaxScaler()),
            ("regressor", cv),
        ]
        pipeline = Pipeline(steps=steps)
        pipeline.fit(X_train, y_train)
        # Register the pipeline and X-target for prediction
        self._lm_dict[name] = (pipeline, X_target)
        # Get train score
        score_train = r2_score(pipeline.predict(X_train), y_train)
        # Get test score
        score_test = r2_score(pipeline.predict(X_test), y_test)
        # Return information regarding regression model
        reg_output = pipeline.named_steps.regressor
        # Intercept
        intercept_df = pd.DataFrame(reg_output.coef_, index=y_train.columns, columns=X_train.columns)
        # Return information
        return {
            **{k: type(v) for (k, v) in steps},
            "alpha": reg_output.alpha_,
            "l1_ratio": reg_output.l1_ratio_,
            "score_train": score_train,
            "score_test": score_test,
            "X_train": X_train,
            "y_train": y_train,
            "X_test": X_test,
            "y_test": y_test,
            "X_target": X_target,
            "intercept": intercept_df,
            "delay": delay
        }

    def predict(self, name="Main"):
        """
        Predict parameter values of the future phases using Elastic Net regression with OxCGRT scores,
        assuming that OxCGRT scores will impact on ODE parameter values with delay.
        New future phases will be added (over-written).

        Args:
            name (str): scenario name

        Raises:
            covsirphy.UnExecutedError: Scenario.fit() was not performed
            NotRegisteredExtraError: no extra datasets were registered

        Returns:
            covsirphy.Scenario: self
        """
        # Arguments
        if name not in self._lm_dict:
            raise UnExecutedError(f"Scenario.fit(name={name})")
        model = self._tracker(name).last_model
        # Prediction with regression model
        pipeline, X_target = self._lm_dict[name]
        predicted = pipeline.predict(X_target)
        # -> end_date/parameter values
        df = pd.DataFrame(predicted, index=X_target.index, columns=model.PARAMETERS)
        df = df.applymap(lambda x: np.around(x, 4 - int(floor(log10(abs(x)))) - 1))
        df.index = [date.strftime(self.DATE_FORMAT) for date in df.index]
        df.index.name = "end_date"
        phase_df = df.drop_duplicates(keep="last").reset_index()
        # Select the last values
        phase_df = phase_df.iloc[[-1], :]
        # Set new future phases
        for phase_dict in phase_df.to_dict(orient="records"):
            self.add(name=name, **phase_dict)
        return self

    def fit_predict(self, oxcgrt_data=None, name="Main", **kwargs):
        """
        Predict parameter values of the future phases using Elastic Net regression with OxCGRT scores,
        assuming that OxCGRT scores will impact on ODE parameter values with delay.
        New future phases will be added (over-written).

        Args:
            oxcgrt_data (covsirphy.OxCGRTData or None): OxCGRT dataset
            name (str): scenario name
            kwargs: the other arguments of Scenario.fit()

        Raises:
            covsirphy.UnExecutedError: Scenario.estimate() or Scenario.add() were not performed
            NotRegisteredExtraError: no extra datasets were registered

        Returns:
            covsirphy.Scenario: self

        Note:
            @oxcgrt_data argument was deprecated. Please use Scenario.register(extras=[oxcgrt_data]).
        """
        self.fit(oxcgrt_data=oxcgrt_data, name=name, **kwargs)
        self.predict(name=name)
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
import numpy as np
import pandas as pd
from scipy.integrate import solve_ivp
from covsirphy.util.error import deprecate
from covsirphy.util.term import Term
from covsirphy.ode.mbase import ModelBase


class ODESimulator(Term):
    """
    Simulation of an ODE model for one phase.

    Args:
        country (str or None): country name
        province (str or None): province name
    """

    def __init__(self, country=None, province=None):
        self.country = country or self.UNKNOWN
        self.province = province or self.UNKNOWN
        # keys: model, step_n, population, param_dict, y0_dict
        self.setting = {}
        # key: non-dim variable name, value: dimensional variable name
        self.var_dict = {}

    def add(self, model, step_n, population, param_dict=None, y0_dict=None):
        """
        Add models to the simulator.

        Args:
            model (subclass of cs.ModelBase): the first ODE model
            step_n (int): the number of steps
            population (int): population in the place
            param_dict (dict):
                - key (str): parameter name
                - value (float): parameter value
                - dictionary of parameter values or None
                - if not include some params, the last values will be used
                    - NameError when the model is the first model
                    - NameError if new params are included
            y0_dict (dict):
                - key (str): variable name
                - value (float): initial value
                - dictionary of dimensional initial values or None
                - None or if not include some variables, the last values will be used
                    - NameError when the model is the first model
                    - NameError if new variable are included
        """
        if self.setting:
            raise ValueError(
                "Simulation for two phases is not supported from version 2.7.0")
        # Register the setting
        self.setting = {
            "model": self._ensure_subclass(model, ModelBase, name="model"),
            self.STEP_N: self._ensure_natural_int(step_n, name="step_n"),
            "population": self._ensure_population(population),
            ModelBase.PARAM_DICT: self._ensure_parameters(model, param_dict),
            self.Y0_DICT: self._ensure_initial_values(model, y0_dict),
        }
        # Update variable dictionary
        self.var_dict.update(model.VAR_DICT)

    def _ensure_parameters(self, model, param_dict):
        """
        Validate the dictionary of parameters.

        Args:
            model (subclass of cs.ModelBase): the ODE model
            param_dict (dict):
                - key (str): parameter name
                - value (float): parameter value

        Returns:
            dict(str, str): dictionary of parameters

        Note:
            If a parameter value is not registered, None will be registered.
        """
        param_dict = param_dict or {}
        usable_dict = {
            p: param_dict[p] if p in param_dict else None for p in model.PARAMETERS}
        if None not in usable_dict.values():
            return usable_dict
        none_params = [k for (k, v) in usable_dict.items() if v is None]
        s = "s" if len(none_params) > 1 else ""
        raise NameError(
            f"Parameter value{s} of {', '.join(none_params)} must be specified by @param_dict."
        )

    def _ensure_initial_values(self, model, y0_dict):
        """
        Validate the dictionary of initial values.

        Args:
            model (subclass of cs.ModelBase): the ODE model
            y0_dict (dict): dictionary of initial values
                - key (str): dimensional variable name
                - value (int):initial value of the variable

        Returns:
            dict(str, str): dictionary of initial values

        Note:
            If initial value of a variable is not registered, None will be registered.
        """
        y0_dict = y0_dict or {}
        usable_dict = {
            v: y0_dict[v] if v in y0_dict else None for v in model.VARIABLES}
        if None not in usable_dict.values():
            return usable_dict
        none_vars = [k for (k, v) in usable_dict.items() if v is None]
        s = "s" if len(none_vars) > 1 else ""
        raise NameError(
            f"Initial value{s} of {', '.join(none_vars)} must be specified by @y0_dict."
        )

    def _solve_ode(self, model, step_n, param_dict, y0_dict, population):
        """
        Solve ODE of the model.

        Args:
            model (subclass of cs.ModelBase): the ODE model
            step_n (int): the number of steps
            param_dict (dict): dictionary of parameter values
                - key (str): parameter name
                - value (float): parameter value
            y0_dict (dict): dictionary of initial values
                - key (str): dimensional variable name
                - value (int):initial value of the variable
            population (int): total population

        Returns:
            (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): Elapsed time divided by tau value [-]
                    - columns with dimensional variables
        """
        tstart, dt, tend = 0, 1, step_n
        variables = model.VARIABLES[:]
        initials = [y0_dict[var] for var in variables]
        sol = solve_ivp(
            fun=model(population=population, **param_dict),
            t_span=[tstart, tend],
            y0=np.array(initials, dtype=np.int64),
            t_eval=np.arange(tstart, tend + dt, dt),
            dense_output=False
        )
        t_df = pd.Series(data=sol["t"], name=self.TS)
        y_df = pd.DataFrame(data=sol["y"].T.copy(), columns=variables)
        y_df = y_df.round()
        return pd.concat([t_df, y_df], axis=1)

    @deprecate(
        old="ODESimulator.run()",
        new="ODESimulator.taufree(), .non_dim() or .dim(tau, start_date) directly")
    def run(self):
        """
        From version 2.7.0, it is not necessary to perform ODESimulator.run().
        Please directory use ODESimulator.taufree(), .non_dim() or .dim(tau, start_date)
        """
        return self.taufree()

    def taufree(self):
        """
        Return tau-free results.

        Returns:
            (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): Elapsed time divided by tau value [-]
                    - columns with dimensionalized variables
        """
        df = self._solve_ode(**self.setting)
        df[self.TS] = df.index
        return df.reset_index(drop=True)

    def non_dim(self):
        """
        Return the non-dimensionalized results.

        Returns:
            (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): Elapsed time divided by tau value [-]
                    - non-dimensionalized variables of Susceptible etc.
        """
        df = self.taufree()
        df = df.set_index(self.TS)
        df = df.apply(lambda x: x / sum(x), axis=1)
        var_dict_rev = {v: k for (k, v) in self.var_dict.items()}
        df.columns = [var_dict_rev[col] for col in df.columns]
        df = df.reset_index()
        return df

    def dim(self, tau, start_date):
        """
        Return the dimensionalized results.

        Args:
            tau (int): tau value [min]
            start_date (str): start date of the records, like 22Jan2020

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pd.TimeStamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - variables of the models (int)
        """
        df = self.taufree()
        df = df.drop(self.TS, axis=1).reset_index(drop=True)
        var_cols = df.columns.tolist()
        df = df.astype(np.int64)
        # Date
        start_obj = datetime.strptime(start_date, self.DATE_FORMAT)
        elapsed = pd.Series(df.index * tau)
        df[self.DATE] = start_obj + elapsed.apply(
            lambda x: timedelta(minutes=x)
        )
        # Place
        df[self.COUNTRY] = self.country
        df[self.PROVINCE] = self.province
        # Return the dataframe
        df = df.loc[:, [self.DATE, self.COUNTRY, self.PROVINCE, *var_cols]]
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime


class StopWatch(object):
    """
    Calculate elapsed time.
    """

    def __init__(self):
        self.start_time = datetime.now()

    def stop(self):
        """
        Stop.

        Returns:
            int: elapsed time [sec]
        """
        end_time = datetime.now()
        return (end_time - self.start_time).total_seconds()

    @staticmethod
    def show(time_sec):
        """
        Show the elapsed time as string.

        Args:
            time_sec (int): time [sec]

        Returns:
            str: eg. '1 min 30 sec'
        """
        minutes, seconds = divmod(int(time_sec), 60)
        return f"{minutes} min {seconds:>2} sec"

    def stop_show(self):
        """
        Stop and show time.

        Returns:
            str: eg. '1 min 30 sec'
        """
        return self.show(self.stop())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import warnings
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter
import numpy as np
import optuna
import pandas as pd
import seaborn as sns
from sklearn.metrics import mean_squared_log_error
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.term import Term
from covsirphy.ode.mbase import ModelBase
from covsirphy.simulation.simulator import ODESimulator


class Estimator(Term):
    """
    Hyperparameter optimization of an ODE model.

    Args:
        record_df (pandas.DataFrame)
            Index
                reset index
            Columns
                - Date (pd.TimeStamp): Observation date
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - any other columns will be ignored
        model (covsirphy.ModelBase): ODE model
        population (int): total population in the place
        tau (int): tau value [min], a divisor of 1440
        kwargs: parameter values of the model and data subseting
    """
    np.seterr(divide="raise")
    optuna.logging.disable_default_handler()
    warnings.simplefilter("ignore", FutureWarning)
    warnings.simplefilter("ignore", SyntaxWarning)

    def __init__(self, record_df, model, population, tau=None, **kwargs):
        # ODE model
        self.model = self._ensure_subclass(model, ModelBase, name="model")
        self.variables = model.VARIABLES[:]
        self.variables_evaluate = [
            v for (v, p) in zip(model.VARIABLES, model.WEIGHTS) if p > 0]
        # Dataset
        if not set(self.NLOC_COLUMNS).issubset(record_df.columns):
            record_df = model.restore(record_df)
        self.record_df = self._ensure_dataframe(
            record_df, name="record_df", columns=self.NLOC_COLUMNS)
        # Settings for simulation
        self.population = self._ensure_population(population)
        df = model.tau_free(self.record_df, population, tau=None)
        self.y0_dict = {
            k: df.loc[df.index[0], k] for k in model.VARIABLES}
        # Fixed parameter values
        self.fixed_dict = {
            k: v for (k, v) in kwargs.items()
            if k in set(model.PARAMETERS) and v is not None
        }
        # For optimization
        self.study = None
        self.total_trials = 0
        self.runtime = 0
        # Tau value
        self.tau_final = self._ensure_tau(tau)
        self.tau_candidates = self.divisors(1440)
        self.tau = tau
        if tau is None:
            self.step_n = None
            self.taufree_df = pd.DataFrame()
        else:
            self._set_taufree()

    def _init_study(self, seed, pruner, upper, percentile):
        """
        Initialize Optuna study.

        Args:
            seed (int or None): random seed of hyperparameter optimization
            pruner (str): Hyperband, Median, Threshold or Percentile
            upper (float): works for "threshold" pruner,
                intermediate RMSLE score is larger than this value, it prunes
            percentile (float): works for "threshold" pruner,
                the best intermediate value is in the bottom percentile among trials, it prunes
        """
        pruner_dict = {
            "hyperband": optuna.pruners.HyperbandPruner(),
            "median": optuna.pruners.MedianPruner(),
            "threshold": optuna.pruners.ThresholdPruner(upper=upper),
            "percentile": optuna.pruners.PercentilePruner(percentile=percentile),
        }
        try:
            self.study = optuna.create_study(
                direction="minimize",
                sampler=optuna.samplers.TPESampler(seed=seed),
                pruner=pruner_dict[pruner.lower()],
            )
        except KeyError:
            raise KeyError(
                f"@pruner should be selected from {', '.join(pruner_dict.keys())}.")

    def run(self, timeout=180, reset_n_max=3, timeout_iteration=10, tail_n=4, allowance=(0.98, 1.02),
            seed=0, pruner="threshold", upper=0.5, percentile=50, **kwargs):
        """
        Run optimization.
        If the result satisfied the following conditions, optimization ends.
        - RMSLE score did not change in the last @tail_n iterations.
        - Monotonic increasing variables increases monotonically.
        - Predicted values are in the allowance when each actual value shows max value.

        Args:
            timeout (int): timeout of optimization
            reset_n_max (int): if study was reset @reset_n_max times, will not be reset anymore
            timeout_iteration (int): time-out of one iteration
            tail_n (int): the number of iterations to decide whether RMSLE score did not change for the last iterations
            allowance (tuple(float, float)): the allowance of the predicted value
            seed (int or None): random seed of hyperparameter optimization
            pruner (str): hyperband, median, threshold or percentile
            upper (float): works for "threshold" pruner,
                intermediate RMSLE score is larger than this value, it prunes
            percentile (float): works for "Percentile" pruner,
                the best intermediate value is in the bottom percentile among trials, it prunes
            kwargs: other keyword arguments will be ignored

        Note:
            @n_jobs was obsoleted because this is not effective for Optuna.
        """
        # Create a study of optuna
        if self.study is None:
            self._init_study(
                seed=seed, pruner=pruner, upper=upper, percentile=percentile)
        reset_n = 0
        iteration_n = math.ceil(timeout / timeout_iteration)
        increasing_cols = [f"{v}{self.P}" for v in self.model.VARS_INCLEASE]
        stopwatch = StopWatch()
        rmsle_scores = []
        for _ in range(iteration_n):
            # Perform optimization
            self.study.optimize(
                self._objective, n_jobs=1, timeout=timeout_iteration)
            # If RMSLE did not change in the last iterations, stop running
            tau, param_dict = self._param()
            rmsle_scores.append(self._rmsle(tau=tau, param_dict=param_dict))
            if len(rmsle_scores) >= tail_n and len(set(rmsle_scores[-tail_n:])) == 1:
                break
            # Create a table to compare observed/estimated values
            comp_df = self._compare(tau=tau, param_dict=param_dict)
            # Check monotonic variables
            mono_ok_list = [
                comp_df[col].is_monotonic_increasing for col in increasing_cols
            ]
            if not all(mono_ok_list):
                if reset_n == reset_n_max - 1:
                    break
                # Initialize the study
                self._init_study(seed=seed)
                reset_n += 1
                continue
            # Need additional trials when the values are not in allowance
            if self._is_in_allowance(comp_df, allowance):
                break
        # Calculate run-time and the number of trials
        self.runtime += stopwatch.stop()
        self.total_trials = len(self.study.trials)

    def _is_in_allowance(self, comp_df, allowance):
        """
        Return whether all max values of predicted values are in allowance or not.

        Args:
            comp_df (pandas.DataFrame): [description]
            allowance (tuple(float, float)): the allowance of the predicted value

        Returns:
            (bool): True when all max values of predicted values are in allowance
        """
        a_max_values = [comp_df[f"{v}{self.A}"].max() for v in self.variables]
        p_max_values = [comp_df[f"{v}{self.P}"].max() for v in self.variables]
        allowance0, allowance1 = allowance
        ok_list = [
            (a * allowance0 <= p) and (p <= a * allowance1)
            for (a, p) in zip(a_max_values, p_max_values)
        ]
        return all(ok_list)

    def _objective(self, trial):
        """
        Objective function of Optuna study.
        This defines the parameter values using Optuna.

        Args:
            trial (optuna.trial): a trial of the study

        Returns:
            float: score of the error function to minimize
        """
        self.tau = self.tau_final or trial.suggest_categorical(
            self.TAU, self.tau_candidates)
        self._set_taufree()
        # Set parameters of the models
        model_param_dict = self.model.param_range(
            self.taufree_df, self.population)
        param_dict = {
            k: self._suggest(trial, k, *v)
            for (k, v) in model_param_dict.items()
            if k not in self.fixed_dict.keys()
        }
        param_dict.update(self.fixed_dict)
        return self._rmsle(self.tau, param_dict)

    def _suggest(self, trial, name, min_value, max_value):
        """
        Suggest parameter value for the trial.

        Args:
            trial (optuna.trial): a trial of the study
            name (str): parameter name
            min_value (float): minimum value of the parameter
            max_value (float): max value of the parameter

        Returns:
            optuna.trial
        """
        try:
            return trial.suggest_uniform(name, min_value, max_value)
        except (OverflowError, np.AxisError):
            return trial.suggest_uniform(name, 0, 1)

    def _set_taufree(self):
        """
        Divide T by tau in the training dataset and calculate the number of steps.
        """
        self.taufree_df = self.model.tau_free(
            self.record_df, self.population, tau=self.tau)
        self.step_n = int(self.taufree_df[self.TS].max())

    def _rmsle(self, tau, param_dict):
        """
        Calculate RMSLE score.

        Args:
            tau (int): tau value [min]
            param_dict (dict[str, int or float]): dictionary of parameter values

        Returns:
            float: RMSLE score
        """
        comp_df = self._compare(tau, param_dict)
        rec_df = comp_df.loc[:, [
            f"{v}{self.A}" for v in self.variables_evaluate]]
        sim_df = comp_df.loc[:, [
            f"{v}{self.P}" for v in self.variables_evaluate]]
        msle = mean_squared_log_error(sim_df, rec_df)
        return np.sqrt(msle)

    def _simulate(self, step_n, param_dict):
        """
        Simulate the values with the parameters.

        Args:
            step_n (int): number of iteration
            dict[str, int or float]: dictionary of parameter values

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - t (int): Elapsed time divided by tau value [-]
                    - columns with dimensionalized variables
        """
        simulator = ODESimulator()
        simulator.add(
            model=self.model,
            step_n=step_n,
            population=self.population,
            param_dict=param_dict,
            y0_dict=self.y0_dict
        )
        return simulator.taufree()

    def _compare(self, tau, param_dict):
        """
        Return comparison table.

        Args:
            tau (int): tau value [min]
            dict[str, int or float]: dictionary of parameter values

        Returns:
            pandas.DataFrame:
                Index
                    t (int): Elapsed time divided by tau value [-]
                Columns
                    - columns with "_actual"
                    - columns with "_predicted:
                    - columns are defined by self.variables
        """
        self.tau = tau
        self._set_taufree()
        sim_df = self._simulate(self.step_n, param_dict)
        df = self.taufree_df.merge(
            sim_df, on=self.TS, suffixes=(self.A, self.P))
        return df.set_index(self.TS)

    def _param(self):
        """
        Return the estimated parameters as a dictionary.

        Returns:
            tuple(int, dict[str, int or float]): tau value and dictionary of parameter values
        """
        try:
            param_dict = self.study.best_params.copy()
        except ValueError:
            param_dict = {p: 0 for p in self.model.PARAMETERS}
            if self.tau_final is None:
                param_dict[self.TAU] = None
        param_dict.update(self.fixed_dict)
        tau = self.tau_final or param_dict.pop(self.TAU)
        return (tau, param_dict)

    def to_dict(self):
        """
        Summarize the results of optimization.

        Returns:
            dict[str, float or int]:
                - (parameters of the model)
                - tau
                - Rt: basic or phase-dependent reproduction number
                - (dimensional parameters [day])
                - RMSLE: Root Mean Squared Log Error
                - Trials: the number of trials
                - Runtime: run time of estimation
        """
        tau, param_dict = self._param()
        model_instance = self.model(population=self.population, **param_dict)
        return {
            **param_dict,
            self.TAU: tau,
            self.RT: model_instance.calc_r0(),
            **model_instance.calc_days_dict(tau),
            self.RMSLE: self._rmsle(tau, param_dict),
            self.TRIALS: self.total_trials,
            self.RUNTIME: StopWatch.show(self.runtime)
        }

    def _history(self):
        """
        Return dataframe to show the history of optimization.

        Args:
            show_figure (bool): if True, show the history as a pair-plot of parameters.
            filename (str): filename of the figure, or None (show figure)

        Returns:
            pandas.DataFrame: the history
        """
        # Create dataframe of the history
        df = self.study.trials_dataframe()
        series = df["datetime_complete"] - df["datetime_start"]
        df["time[s]"] = series.dt.total_seconds()
        drop_cols = [
            "datetime_complete", "datetime_start", "system_attrs__number"]
        return df.drop(drop_cols, axis=1, errors="ignore")

    def history(self, show_figure=True, filename=None):
        """
        Return dataframe to show the history of optimization.

        Args:
            show_figure (bool): if True, show the history as a pair-plot of parameters.
            filename (str): filename of the figure, or None (show figure)

        Returns:
            pandas.DataFrame: the history
        """
        df = self._history()
        if not show_figure:
            return df
        # Show figure
        fig_df = df.loc[:, df.columns.str.startswith("params_")]
        fig_df.columns = fig_df.columns.str.replace("params_", "")
        warnings.simplefilter("ignore", category=UserWarning)
        sns.pairplot(fig_df, diag_kind="kde", markers="+")
        # Save figure or show figure
        if filename is None:
            plt.show()
            return df
        plt.savefig(filename, bbox_inches="tight", transparent=False, dpi=300)
        plt.clf()
        return df

    def accuracy(self, show_figure=True, filename=None):
        """
        Show accuracy as a figure.

        Args:
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (show figure)

        Returns:
            pandas.DataFrame:
                Index
                    t (int): Elapsed time divided by tau value [-]
                Columns
                    - columns with "_actual"
                    - columns with "_predicted:
                    - columns are defined by self.variables
        """
        # Create a table to compare observed/estimated values
        df = self._compare(*self._param())
        if not show_figure:
            return df
        # Variables to show accuracy
        variables = [
            v for (i, (p, v))
            in enumerate(zip(self.model.WEIGHTS, self.model.VARIABLES))
            if p != 0 and i != 0
        ]
        # Prepare figure object
        val_len = len(variables) + 1
        fig, axes = plt.subplots(
            ncols=1, nrows=val_len, figsize=(9, 6 * val_len / 2))
        # Comparison of each variable
        for (ax, v) in zip(axes.ravel()[1:], variables):
            df[[f"{v}{self.A}", f"{v}{self.P}"]].plot.line(
                ax=ax, ylim=(None, None), sharex=True,
                title=f"{self.model.NAME}: Comparison regarding {v}(t)"
            )
            ax.yaxis.set_major_formatter(ScalarFormatter(useMathText=True))
            ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
            ax.legend(
                bbox_to_anchor=(1.02, 0), loc="lower left", borderaxespad=0
            )
        # Summarize in a figure
        for v in variables:
            df[f"{v}_diff"] = df[f"{v}{self.A}"] - df[f"{v}{self.P}"]
            df[f"{v}_diff"].plot.line(
                ax=axes.ravel()[0], sharex=True,
                title=f"{self.model.NAME}: observed - estimated"
            )
        axes.ravel()[0].axhline(y=0, color="black", linestyle="--")
        axes.ravel()[0].yaxis.set_major_formatter(
            ScalarFormatter(useMathText=True))
        axes.ravel()[0].ticklabel_format(
            style="sci", axis="y", scilimits=(0, 0))
        axes.ravel()[0].legend(
            bbox_to_anchor=(1.02, 0), loc="lower left", borderaxespad=0)
        fig.tight_layout()
        # Save figure or show figure
        if filename is None:
            plt.show()
            return df
        plt.savefig(filename, bbox_inches="tight", transparent=False, dpi=300)
        plt.clf()
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import matplotlib
import japanmap
import pandas as pd
from covsirphy.util.error import deprecate


@deprecate(old="covsirphy.jpn_map", new="JHUData.map('Japan')", version="2.16")
def jpn_map(prefectures, values, title, cmap_name="Reds", filename=None):
    """
    Show colored Japan prefecture map.

    Args:
        prefectures (list[str] or pd.Series[str]): prefecture name.
        values (int or float): value of each prefectures
        title (str): title of the figure
        cmap_name (str): Please refere to
            https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
        filename (str): filename of the figure, or None (show figure)
    """
    # Prefecture code created in
    # https://www.kaggle.com/lisphilar/eda-of-japan-dataset
    # Primary data: http://nlftp.mlit.go.jp/ksj/gml/codelist/PrefCd.html
    # cf. https://www.japanvisitor.com/japan-travel/prefectures-map
    pref_code_dict = dict(
        [
            ('Hokkaido', 1), ('Aomori', 2), ('Iwate', 3),
            ('Miyagi', 4), ('Akita', 5), ('Yamagata', 6),
            ('Fukushima', 7), ('Ibaraki', 8), ('Tochigi', 9),
            ('Gunma', 10), ('Saitama', 11), ('Chiba', 12),
            ('Tokyo', 13), ('Kanagawa', 14), ('Niigata', 15),
            ('Toyama', 16), ('Ishikawa', 17), ('Fukui', 18),
            ('Yamanashi', 19), ('Nagano', 20), ('Gifu', 21),
            ('Shizuoka', 22), ('Aichi', 23), ('Mie', 24),
            ('Shiga', 25), ('Kyoto', 26), ('Osaka', 27), ('Hyogo', 28),
            ('Nara', 29), ('Wakayama', 30), ('Tottori', 31),
            ('Shimane', 32), ('Okayama', 33), ('Hiroshima', 34),
            ('Yamaguchi', 35), ('Tokushima', 36), ('Kagawa', 37),
            ('Ehime', 38), ('Kochi', 39), ('Fukuoka', 40),
            ('Saga', 41), ('Nagasaki', 42), ('Kumamoto', 43),
            ('Oita', 44), ('Miyazaki', 45), ('Kagoshima', 46),
            ('Okinawa', 47)
        ]
    )
    # Data to dataframe
    df = pd.DataFrame({"Name": prefectures, "Value": values})
    df["Code"] = df["Name"].map(pref_code_dict)
    df = df.dropna().reset_index()
    df.index = df["Code"].apply(lambda x: japanmap.pref_names[int(x)])
    # Color code
    cmap = matplotlib.cm.get_cmap(cmap_name)
    norm = matplotlib.colors.Normalize(
        vmin=df["Value"].min(), vmax=df["Value"].max())
    # Show figure
    plt.xticks(color="None")
    plt.yticks(color="None")
    plt.tick_params(length=0)
    plt.imshow(
        japanmap.picture(df["Value"].apply(
            lambda x: "#" + bytes(cmap(norm(x), bytes=True)[:3]).hex()
        )))
    mappable = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)
    mappable._A = []
    plt.colorbar(mappable)
    plt.title(title)
    if filename is None:
        plt.show()
        return None
    plt.savefig(
        filename, bbox_inches="tight", transparent=False, dpi=300
    )
    plt.clf()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import numpy as np
import pandas as pd
from covsirphy.util.error import UnExecutedError
from covsirphy.cleaning.cbase import CleaningBase


class CountryData(CleaningBase):
    """
    Data cleaning of country level data.

    Args:
        filename (str or None): filename to read the data
        country (str): country name
        province (str or None): province name

    Note:
        If province name will be set in CountryData.set_variables(), @province will be ignored.
    """

    def __init__(self, filename, country, province=None):
        self._raw = pd.DataFrame() if filename is None else self.load(filename)
        self._country = country
        self._province = province
        self.province_col = None
        self.var_dict = {}
        self._cleaned_df = pd.DataFrame()
        self._citation = str()
        # Directory that save the file
        if filename is None:
            self._dirpath = Path("input")
        else:
            self._dirpath = Path(filename).resolve().parent

    @property
    def country(self):
        """
        str: country name
        """
        return self._country

    def raw_columns(self):
        """
        Return the column names of the raw data.

        Returns:
            list[str]: the list of column names of the raw data
        """
        return self._raw.columns.tolist()

    def set_variables(self, date, confirmed, fatal, recovered, province=None):
        """
        Set the correspondence of the variables and columns of the raw data.

        Args:
            date (str): column name for Date
            confirmed (str): column name for Confirmed
            fatal (str): column name for Fatal
            recovered (str): column name for Confirmed
            province (str): (optional) column name for Province
        """
        self.province_col = province
        self.var_dict = {
            date: self.DATE,
            confirmed: self.C,
            fatal: self.F,
            recovered: self.R
        }

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.
        This method overwrite super()._cleaning() method.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.TimeStamp): Observation date
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        if not self.var_dict:
            raise UnExecutedError("CountryData.set_variables()")
        df = self._raw.copy()
        # Rename the columns
        df = df.rename(self.var_dict, axis=1)
        # Confirm the expected columns are in raw data
        expected_cols = [
            self.DATE, self.C, self.F, self.R
        ]
        self._ensure_dataframe(df, name="the raw data", columns=expected_cols)
        # Remove empty rows
        df = df.dropna(subset=[self.DATE])
        # Add province column
        if self.province_col is not None:
            df = df.rename({self.province_col: self.PROVINCE}, axis=1)
        else:
            df[self.PROVINCE] = self._province or self.UNKNOWN
        # Values
        v_cols = [self.C, self.F, self.R]
        for col in v_cols:
            df[col] = pd.to_numeric(df[col], errors="coerce")
        df[v_cols] = df[v_cols].fillna(0).astype(np.int64)
        df[self.CI] = df[self.C] - df[self.F] - df[self.R]
        # Groupby date and province
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        df = df.groupby([self.DATE, self.PROVINCE]).sum().reset_index()
        # Add country column
        df[self.COUNTRY] = self._country
        df = df.loc[:, self.COLUMNS]
        # Update data types to reduce memory
        df[self.AREA_COLUMNS] = df[self.AREA_COLUMNS].astype("category")
        return df

    def cleaned(self):
        """
        Return the cleaned dataset.
        Cleaning method is defined by CountryData._cleaning() method.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.TimeStamp): Observation date
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/sstate name
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        if self._cleaned_df.empty:
            self._cleaned_df = self._cleaning()
        return self._cleaned_df

    def total(self):
        """
        Return a dataframe to show chronological change of number and rates.

        Returns:
            pandas.DataFrame: group-by Date, sum of the values

                Index
                    Date (pd.TimeStamp): Observation date
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Fatal per Confirmed (int)
                    - Recovered per Confirmed (int)
                    - Fatal per (Fatal or Recovered) (int)
        """
        df = self.cleaned()
        # Calculate total values at country level if not registered
        c_level_df = df.groupby(self.DATE).sum().reset_index()
        c_level_df[self.PROVINCE] = self.UNKNOWN
        df = pd.concat([df, c_level_df], axis=0, ignore_index=True)
        df = df.drop_duplicates(subset=[self.DATE, self.PROVINCE])
        df = df.loc[df[self.PROVINCE] == self.UNKNOWN, :]
        df = df.drop([self.COUNTRY, self.PROVINCE], axis=1)
        df = df.set_index(self.DATE)
        # Calculate rates
        total_series = df.sum(axis=1)
        r_cols = self.RATE_COLUMNS[:]
        df[r_cols[0]] = df[self.F] / total_series
        df[r_cols[1]] = df[self.R] / total_series
        df[r_cols[2]] = df[self.F] / (df[self.F] + df[self.R])
        return df.loc[:, [*self.VALUE_COLUMNS, *r_cols]]

    def countries(self):
        """
        Return names of countries where records are registered.

        Returns:
            list[str]: list of country names
        """
        return [self._country]

    def register_total(self):
        """
        Register total value of all provinces as country level data.

        Returns:
            covsirphy.CountryData: self

        Note:
            If country level data was registered, this will be overwritten.
        """
        # Calculate total values at province level
        clean_df = self.cleaned()
        clean_df = clean_df.loc[clean_df[self.PROVINCE] != self.UNKNOWN]
        total_df = clean_df.groupby(self.DATE).sum().reset_index()
        total_df[self.COUNTRY] = self._country
        total_df[self.PROVINCE] = self.UNKNOWN
        # Add/overwrite country level data
        df = clean_df.loc[clean_df[self.PROVINCE] != self.UNKNOWN]
        df = pd.concat([df, total_df], ignore_index=True, sort=True)
        df[self.STR_COLUMNS] = df[self.STR_COLUMNS].astype("category")
        self._cleaned_df = df.loc[:, self.COLUMNS]
        return self

    def map(self, country=None, variable="Confirmed", date=None, **kwargs):
        """
        Create colored map to show the values.

        Args:
            country (None): None
            variable (str): variable name to show
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Raises:
            NotImplementedError: @country was specified
        """
        if country is not None:
            raise NotImplementedError("@country cannot be specified, always None.")
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        title = f"{self._country}: the number of {variable.lower()} cases on {date_str}"
        # Country-specific map
        return self._colored_map_country(
            country=self._country, variable=variable, title=title, date=date, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
try:
    import better_exceptions
    better_exceptions_installed = True
except ImportError:
    better_exceptions_installed = False
# version
from covsirphy.__version__ import __version__
# util
from covsirphy.util.plotting import line_plot, line_plot_multiple
from covsirphy.util.optimize import Optimizer
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.error import deprecate
from covsirphy.util.error import SubsetNotFoundError, ScenarioNotFoundError, UnExecutedError
from covsirphy.util.error import PCRIncorrectPreconditionError, NotInteractiveError, UnExpectedValueError
from covsirphy.util.error import NotRegisteredMainError, NotRegisteredExtraError
from covsirphy.util.file import save_dataframe
from covsirphy.util.argument import find_args
from covsirphy.util.term import Term, Word
# visualization
from covsirphy.visualization.vbase import VisualizeBase
from covsirphy.visualization.colored_map import ColoredMap
from covsirphy.visualization.japan_map import jpn_map
# cleaning
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.jhu_complement import JHUDataComplementHandler
from covsirphy.cleaning.country_data import CountryData
from covsirphy.cleaning.japan_data import JapanData
from covsirphy.cleaning.population import PopulationData, Population
from covsirphy.cleaning.pyramid import PopulationPyramidData
from covsirphy.cleaning.oxcgrt import OxCGRTData
from covsirphy.cleaning.pcr_data import PCRData
from covsirphy.cleaning.covid19datahub import COVID19DataHub
from covsirphy.cleaning.linelist import LinelistData
from covsirphy.cleaning.vaccine_data import VaccineData
from covsirphy.cleaning.dataloader import DataLoader
# ode
from covsirphy.ode.mbase import ModelBase
from covsirphy.ode.sir import SIR
from covsirphy.ode.sird import SIRD
from covsirphy.ode.sirf import SIRF
from covsirphy.ode.sirfv import SIRFV
from covsirphy.ode.sewirf import SEWIRF
# simulation
from covsirphy.simulation.estimator import Estimator
from covsirphy.simulation.simulator import ODESimulator
# phase
from covsirphy.phase.trend import Trend
from covsirphy.phase.sr_change import ChangeFinder
from covsirphy.phase.phase_unit import PhaseUnit
from covsirphy.phase.phase_series import PhaseSeries
from covsirphy.phase.phase_estimator import MPEstimator
# analysis
from covsirphy.analysis.example_data import ExampleData
from covsirphy.analysis.data_handler import DataHandler
from covsirphy.analysis.param_tracker import ParamTracker
from covsirphy.analysis.scenario import Scenario
from covsirphy.analysis.model_validator import ModelValidator
# worldwide
from covsirphy.worldwide.policy import PolicyMeasures


def get_version():
    """
    Return the version number, like CovsirPhy v0.0.0
    """
    return f"CovsirPhy v{__version__}"


__all__ = [
    "ExampleData", "Scenario", "ModelValidator", "ParamTracker",
    "ODESimulator", "ChangeFinder", "DataHandler",
    "PhaseSeries", "PhaseUnit", "MPEstimator",
    "Term", "CleaningBase", "DataLoader", "COVID19DataHub",
    "JHUData", "CountryData", "PopulationData", "OxCGRTData", "VaccineData",
    "PopulationPyramidData",
    "LinelistData", "PCRData", "JapanData", "JHUDataComplementHandler",
    "ModelBase", "SIR", "SIRD", "SIRF", "SIRFV", "SEWIRF",
    "Estimator", "Trend", "Optimizer",
    "line_plot", "StopWatch", "deprecate", "find_args",
    "line_plot_multiple",
    "save_dataframe",
    "PolicyMeasures",
    "SubsetNotFoundError", "ScenarioNotFoundError", "UnExecutedError",
    "PCRIncorrectPreconditionError", "NotInteractiveError",
    "UnExpectedValueError", "NotRegisteredMainError", "NotRegisteredExtraError",
    "VisualizeBase", "ColoredMap",
    # Deprecated
    "Population", "Word", "jpn_map",
]

# Show exceptions in better format if used from command line
if not hasattr(sys, "ps1") or not better_exceptions_installed:
    better_exceptions.MAX_LENGTH = None
    better_exceptions.hook()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import matplotlib
if not hasattr(sys, "ps1"):
    matplotlib.use("Agg")
from matplotlib import pyplot as plt
from covsirphy.util.term import Term

# Style of Matplotlib
plt.style.use("fast")
plt.rcParams["xtick.direction"] = "in"
plt.rcParams["ytick.direction"] = "in"
plt.rcParams["font.size"] = 11.0
plt.rcParams["figure.figsize"] = (9, 6)
plt.rcParams["legend.frameon"] = False


class VisualizeBase(Term):
    """
    Base class for visualization.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        _, self._ax = plt.subplots(1, 1)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        # Settings
        if self._title:
            self._ax.title.set_text(self._title)
        # Display the figure if filename is None after plotting
        try:
            if self._filename is None:
                plt.show()
            else:
                # Save the image as a file
                plt.savefig(self._filename, **self._savefig_dict)
                plt.clf()
                plt.close("all")
        except AttributeError:
            pass

    @property
    def title(self):
        """
        str: title of the figure
        """
        return self._title

    @title.setter
    def title(self, title):
        self._title = str(title)

    @property
    def ax(self):
        """
        matplotlib.axis: axis
        """
        return self._ax

    @ax.setter
    def ax(self, ax):
        self._ax = self._ensure_instance(ax, matplotlib.axes.Axes, name="ax")

    def plot(self):
        """
        Method for plotting. This will be defined in child classes.

        Raises:
            NotImplementedError: not implemented
        """
        raise NotImplementedError

    def tick_params(self, **kwargs):
        """
        Directly calling matplotlib.pyplot.tick_params,
        change the appearance of ticks, tick labels and gridlines.

        Args:
            kwargs: arguments of matplotlib.pyplot.tick_params
        """
        self._ax.tick_params(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
from pathlib import Path
from dask import dataframe as dd
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate, SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class PopulationData(CleaningBase):
    """
    Data cleaning of total population dataset.

    Args:
        filename (str): CSV filename of the dataset
        citation (str): citation
    """
    POPULATION_COLS = [
        CleaningBase.ISO3,
        CleaningBase.COUNTRY,
        CleaningBase.PROVINCE,
        CleaningBase.DATE,
        CleaningBase.N
    ]

    def __init__(self, filename=None, citation=None):
        self.created_time = datetime.now()
        if filename is None:
            self._raw = pd.DataFrame()
            self._cleaned_df = pd.DataFrame(columns=self.POPULATION_COLS)
        else:
            self._raw = dd.read_csv(
                filename, dtype={"Province/State": "object"}
            ).compute()
            self._cleaned_df = self._cleaning()
        self._citation = citation or ""
        # Directory that save the file
        if filename is None:
            self._dirpath = Path("input")
        else:
            self._dirpath = Path(filename).resolve().parent

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - ISO3 (str): ISO3 code or "-"
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Date (pd.TimeStamp): date of the records (if available) or today
                    - Population (int): total population
        """
        df = self._raw.copy()
        # Rename the columns
        df = df.rename(
            {
                "Country.Region": self.COUNTRY,
                "Country/Region": self.COUNTRY,
                "Province.State": self.PROVINCE,
                "Province/State": self.PROVINCE,
                "ObservationDate": self.DATE
            },
            axis=1
        )
        # Confirm the expected columns are in raw data
        expected_cols = [
            self.COUNTRY, self.PROVINCE, self.N
        ]
        self._ensure_dataframe(df, name="the raw data", columns=expected_cols)
        # ISO3
        df[self.ISO3] = df[self.ISO3] if self.ISO3 in df.columns else self.UNKNOWN
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Country
        df[self.COUNTRY] = df[self.COUNTRY].replace(
            {
                # COD
                "Congo, the Democratic Republic of the": "Democratic Republic of the Congo",
                # COG
                "Congo": "Republic of the Congo",
                # South Korea
                "Korea, South": "South Korea",
            }
        )
        # Province
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.UNKNOWN)
        # Set 'Others' as the country name of cruise ships
        ships = ["Diamond Princess", "Costa Atlantica", "Grand Princess", "MS Zaandam"]
        for ship in ships:
            df.loc[df[self.COUNTRY] == ship, [self.COUNTRY, self.PROVINCE]] = [self.OTHERS, ship]
        # Values
        df = df.dropna(subset=[self.N]).reset_index(drop=True)
        df[self.N] = df[self.N].astype(np.int64)
        # Columns to use
        df = df.loc[
            :, [self.ISO3, self.COUNTRY, self.PROVINCE, self.DATE, self.N]]
        # Remove duplicates
        df = df.drop_duplicates().reset_index(drop=True)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        return df

    def _created_date(self):
        """
        Return 0:00 AM of the created date.

        Returns:
            datetime.datetime
        """
        return self.created_time.replace(hour=0, minute=0, second=0, microsecond=0)

    def total(self):
        """
        Return the total value of population in the dataset.

        Returns:
            int
        """
        values = self._cleaned_df[self.N]
        return int(sum(values))

    def to_dict(self, country_level=True):
        """
        Return dictionary of population values.

        Args:
        country_level (str): whether key is country name or not

        Returns:
            dict
                - if @country_level is True, {"country", population}
                - if False, {"country/province", population}
        """
        df = self._cleaned_df.copy()
        if country_level:
            df = df.loc[df[self.PROVINCE] == self.UNKNOWN, :]
            df["key"] = df[self.COUNTRY]
        else:
            df = df.loc[df[self.PROVINCE] != self.UNKNOWN, :]
            df["key"] = df[self.COUNTRY].str.cat(
                df[self.PROVINCE], sep=self.SEP
            )
        return df.set_index("key").to_dict()[self.N]

    def value(self, country, province=None, date=None):
        """
        Return the value of population in the place.

        Args:
            country (str): country name or ISO3 code
            province (str): province name
            date (str or None): observation date, like 01Jun2020

        Returns:
            int: population in the place

        Note:
            If @date is None, the created date of the instancewill be used
        """
        country_alias = self.ensure_country_name(country)
        try:
            df = self.subset(
                country=country, province=province, start_date=date, end_date=date)
        except KeyError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province, date=date)
        df = df.sort_values(self.DATE)
        return int(df.loc[df.index[-1], [self.N]].values[0])

    def update(self, value, country, province=None, date=None):
        """
        Update the value of a new place.

        Args:
            value (int): population in the place
            country (str): country name
            province (str): province name
            date (str or None): observation date, like 01Jun2020

        Returns:
            covsirphy.PopulationData: self

        Note:
            If @date is None, the created date of the instance will be used.
            If @province is None, "-" will be used.
        """
        population = self._ensure_natural_int(value, "value")
        province = province or self.UNKNOWN
        date_stamp = pd.to_datetime(date or self._created_date())
        df = self._cleaned_df.copy()
        c_series = df[self.COUNTRY]
        p_series = df[self.PROVINCE]
        d_series = df[self.DATE]
        sel = (c_series == country) & (
            p_series == province) & (d_series == date_stamp)
        if not df.loc[sel, :].empty:
            df.loc[sel, self.N] = value
            self._cleaned_df = df.copy()
            return self
        series = pd.Series(
            [self.UNKNOWN, country, province, date_stamp, population],
            index=[self.ISO3, self.COUNTRY, self.PROVINCE, self.DATE, self.N]
        )
        self._cleaned_df = df.append(series, ignore_index=True)
        return self

    def countries(self):
        """
        Return names of countries where records are registered.

        Raises:
            KeyError: Country names are not registered in this dataset

        Returns:
            list[str]: list of country names

        Note:
            Country 'Others' will be removed.
        """
        country_list = super().countries()
        removed_countries = ["Others"]
        country_list = list(set(country_list) - set(removed_countries))
        return country_list

    def map(self, country=None, variable="Population", date=None, **kwargs):
        """
        Create colored map with the number of tests.

        Args:
            country (str or None): country name or None (global map)
            variable (str): always 'Population'
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Raises:
            NotImplementedError: @variable was specified

        Note:
            When @country is None, country level data will be shown on global map.
            When @country is a country name, province level data will be shown on country map.
        """
        if variable != self.N:
            raise NotImplementedError(f"@variable cannot be changed, always {self.N}.")
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        country_str = country or "Global"
        title = f"{country_str}: {variable.lower()} on {date_str}"
        # Global map
        if country is None:
            return self._colored_map_global(
                variable=variable, title=title, date=date, **kwargs)
        # Country-specific map
        return self._colored_map_country(
            country=country, variable=variable, title=title, date=date, **kwargs)


class Population(PopulationData):
    @deprecate(old="Population()", new="PopulationData()")
    def __init__(self, filename):
        super().__init__(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import numpy as np
import pandas as pd
from covsirphy.util.error import SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class VaccineData(CleaningBase):
    """
    Dataset regarding vaccination retrieved from "Our World In Data".
    https://github.com/owid/covid-19-data/tree/master/public/data
    https://ourworldindata.org/coronavirus

    Args:
        filename (str or pathlib.path): CSV filename to save the raw dataset
        force (bool): if True, always download the dataset from the server
        verbose (int): level of verbosity

    Note:
        Columns of VaccineData.cleaned():
            - Date (pandas.TimeStamp): observation dates
            - Country (pandas.Category): country (or province) names
            - ISO3 (pandas.Category): ISO3 codes
            - Product (pandas.Category): product names
            - Vaccinations (int): cumulative number of vaccinations
            - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
            - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
    """
    # URL
    URL = "https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/vaccinations/"
    URL_REC = f"{URL}vaccinations.csv"
    URL_LOC = f"{URL}locations.csv"
    # Columns
    VAC_COLS = [
        CleaningBase.DATE, CleaningBase.COUNTRY, CleaningBase.ISO3, CleaningBase.PRODUCT,
        CleaningBase.VAC, CleaningBase.V_ONCE, CleaningBase.V_FULL]
    VAC_SUBSET_COLS = [CleaningBase.DATE, CleaningBase.VAC, CleaningBase.V_ONCE, CleaningBase.V_FULL]

    def __init__(self, filename, force=False, verbose=1):
        Path(filename).parent.mkdir(exist_ok=True, parents=True)
        if Path(filename).exists() and not force:
            try:
                self._raw = self.load(filename)
            except KeyError:
                # Error when the local dataset does not have necessary columns
                # Raised when new CovsirPhy version requires additional columns
                self._raw = self._retrieve(filename=filename, verbose=verbose)
        else:
            self._raw = self._retrieve(filename=filename, verbose=verbose)
        self._cleaned_df = self._cleaning()
        self._citation = "Hasell, J., Mathieu, E., Beltekian, D. et al." \
            " A cross-country database of COVID-19 testing. Sci Data 7, 345 (2020)." \
            " https://doi.org/10.1038/s41597-020-00688-8"
        # Directory that save the file
        self._dirpath = Path(filename or "input").resolve().parent

    def _retrieve(self, filename, verbose=1):
        """
        Retrieve the dataset from server.
        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            verbose (int): level of verbosity
        Returns:
            pd.DataFrame:
                Index reset index
                Columns Date, Country, Product, Vaccinations
        """
        # Show URL
        if verbose:
            print("Retrieving COVID-19 vaccination dataset from https://github.com/owid/covid-19-data/")
        # Download datasets and merge them
        rename_dict = {
            "date": self.DATE, "location": self.COUNTRY, "iso_code": self.ISO3,
            "vaccines": self.PRODUCT, "total_vaccinations": self.VAC,
            "people_vaccinated": self.V_ONCE,
            "people_fully_vaccinated": self.V_FULL,
        }
        rec_df = self.load(self.URL_REC, columns=list(set(rename_dict) - set(["vaccines"])))
        loc_df = self.load(self.URL_LOC, columns=["location", "vaccines"])
        df = rec_df.merge(loc_df, how="left", on="location")
        df = df.rename(rename_dict, axis=1)
        # Save the dataframe as CSV file
        df.to_csv(filename, index=False)
        return df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                - Date (pandas.TimeStamp): observation dates
                - Country (pandas.Category): country (or province) names
                - ISO3 (pandas.Category): ISO3 codes
                - Product (pandas.Category): product names
                - Vaccinations (int): cumulative number of vaccinations
                - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        df = self._raw.copy()
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        for col in [self.COUNTRY, self.ISO3, self.PRODUCT]:
            df[col] = df[col].astype("category")
        # Fill in NA values
        for col in [self.VAC, self.V_ONCE, self.V_FULL]:
            df[col] = pd.to_numeric(df[col], errors="coerce")
            df[col] = df.groupby(self.ISO3)[col].fillna(method="ffill").fillna(0).astype(np.int64)
        return df.loc[:, self.VAC_COLS]

    def subset(self, country, product=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str or None): country name or ISO3 code
            product (str or None): product name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
                    - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        df = self._cleaned_df.copy()
        # Subset by country
        country_alias = self.ensure_country_name(country)
        df = df.loc[df[self.COUNTRY] == country_alias]
        # Subset by product name
        if product is not None:
            df = df.loc[df[self.PRODUCT] == product]
        # Subset with start date
        if start_date is not None:
            df = df.loc[df[self.DATE] >= self.date_obj(start_date)]
        # Subset with end date
        if end_date is not None:
            df = df.loc[df[self.DATE] <= self.date_obj(end_date)]
        # Check records were found
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=product,
                start_date=start_date, end_date=end_date)
        return df.loc[:, self.VAC_SUBSET_COLS].reset_index(drop=True)

    def records(self, country, product=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str or None): country name or ISO3 code
            product (str or None): product name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
                    - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        return self.subset(
            country=country, product=product, start_date=start_date, end_date=end_date)

    def total(self):
        """
        Calculate total values of the cleaned dataset.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.TimeStamp): observation date
                    - Vaccinations (int): the number of vaccinations
                    - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
        """
        df = self._cleaned_df.copy()
        # Select 'World' data
        df = df.loc[df[self.COUNTRY] == "World"]
        # Resampling
        df = df.set_index(self.DATE).resample("D").sum()
        return df.reset_index()

    def map(self, country=None, variable="Vaccinations", date=None, **kwargs):
        """
        Create colored map with the number of vaccinations.

        Args:
            country (None): always None
            variable (str): variable to show
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Raises:
            NotImplementedError: @country was specified
        """
        if country is not None:
            raise NotImplementedError("@country cannot be specified, always None.")
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        country_str = "Global"
        title = f"{country_str}: the number of {variable.lower()} on {date_str}"
        # Global map
        return self._colored_map_global(variable=variable, title=title, date=date, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from functools import partial
from pathlib import Path
import warnings
import country_converter as coco
import geopandas as gpd
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import requests
from unidecode import unidecode
from covsirphy.visualization.vbase import VisualizeBase


class ColoredMap(VisualizeBase):
    """
    Create global map with pandas.DataFrame.

    Args:
        filename (str or None): filename to save the figure or None (display)
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    def __init__(self, filename=None, **kwargs):
        super().__init__(filename=filename, **kwargs)
        self._to_iso3 = partial(coco.convert, to="ISO3", not_found=None)
        self._geo_dirpath = Path("input")

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    @property
    def directory(self):
        """
        str: directory to save the downloaded files of geometry information
        """
        return str(self._geo_dirpath)

    @directory.setter
    def directory(self, name):
        self._geo_dirpath = Path(name)

    def plot(self, data, level="Country", included=None, excluded=None, logscale=True, **kwargs):
        """
        Set dataframe and the variable to show in a colored map.

        Args:
            data (pandas.DataFrame): data to show
                Index
                    reset index
                Columns
                    - Country (str or pandas.Category): country name(s)
                    - Province (str or pandas.Category): province names, necessary when @level is 'Province'
                    - Value (int or float or None): values to coloring the map
                    - ISO3 (str): ISO3 codes, optional
            level (str): 'Country' (global map) or 'Province' (country-specific map)
            logscale (bool): whether convert the value to log10 scale values or not
            included (list[str] or None): included countries/provinces or None (all)
            excluded (list[str] or None): excluded countries/provinces or None (all)
            kwargs: arguments of geopandas.GeoDataFrame.plot() except for 'column'

        Raises:
            ValueError: labels for data are not unique
            UnExpectedValueError: some countries' records are included when @level is 'Province'
            SubsetNotFoundError: no geometry information available for the labels
        """
        expected_cols = [self.COUNTRY, "Value"] + [] if level == self.COUNTRY else [self.PROVINCE]
        self._ensure_dataframe(data, name="data", columns=expected_cols)
        if level == self.COUNTRY:
            # Global map with country level data
            if not data[self.COUNTRY].is_unique:
                raise ValueError(f"{self.COUNTRY} column of data should be unique.")
            gdf = self._global_data(data=data, included=included, excluded=excluded)
        else:
            # Country-specific map with province level data
            unique_n = data[self.COUNTRY].nunique()
            if unique_n > 1:
                raise ValueError(
                    f"{self.COUNTRY} column of data should have only one country name when @level is {self.PROVINCE}, but {unique_n} found.")
            country = data[self.COUNTRY].value_counts().index[0]
            if not data[self.PROVINCE].is_unique:
                raise ValueError(f"{self.PROVINCE} column of data should be unique.")
            gdf = self._country_specific_data(data, included=included, excluded=excluded, country=country)
        gdf.loc[gdf["Value"] < 0, "Value"] = 0
        # Colorbar
        divider = make_axes_locatable(self._ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        # Arguments of plotting with GeoPandas
        plot_kwargs = {
            "legend": True,
            "cmap": "coolwarm",
            "ax": self._ax,
            "cax": cax,
            "missing_kwds": {
                "color": "lightgrey",
                "edgecolor": "white",
                "hatch": "///",
            }
        }
        plot_kwargs.update(kwargs)
        # Convert to log10 scale
        if logscale:
            gdf["Value"] = np.log10(gdf["Value"] + 1)
            plot_kwargs["legend_kwds"] = {"label": "in log10 scale"}
        # Plotting
        warnings.filterwarnings("ignore", category=UserWarning)
        if not gdf["Value"].isna().sum():
            # Missing values are not included
            plot_kwargs.pop("missing_kwds")
        gdf.plot(column="Value", **plot_kwargs)
        # Remove all ticks
        self._ax.tick_params(labelbottom=False, labelleft=False, left=False, bottom=False)

    def _global_data(self, data, included, excluded):
        """
        Create global map data with geometry information.

        Args:
            data (pandas.DataFrame): data to show
                Index
                    reset index
                Columns
                    - Country (str): country names
                    - Value (int or float or None): values to plot
                    - ISO3 (str): ISO3 codes, optional
            included (list[str] or None): included countries or None (all)
            excluded (list[str] or None): excluded countries or None (all)

        Returns:
            geopandas.GeoDataFrame:
                Index
                    reset index
                Columns
                    - Value (int or float or None)
                    - geometry (geopandas.GeoDataFrame.geometry): geometry information
        """
        # data to plot
        df = data.copy()
        df[self.ISO3] = df[self.COUNTRY].apply(self._to_iso3) if self.ISO3 not in df else df[self.ISO3]
        # Geometry
        gdf = self._load_geo_global()
        # Merge them
        gdf = gdf.merge(df, how="left", on=self.ISO3)
        # Select countries
        included_codes = gdf[self.ISO3].tolist() if included is None else [self._to_iso3(c) for c in included]
        excluded_codes = [] if excluded is None else [self._to_iso3(c) for c in excluded]
        sel = set(included_codes) - set(excluded_codes)
        return gdf.loc[gdf[self.ISO3].isin(sel), ["Value", "geometry"]]

    def _country_specific_data(self, data, included, excluded, country):
        """
        Create country-specific map data with geometry information.

        Args:
            data (pandas.DataFrame): data to show
                Index
                    reset index
                Columns
                    - Province (str): province names
                    - Value (int or float or None): values to plot
            included (list[str] or None): included countries or None (all)
            excluded (list[str] or None): excluded countries or None (all)
            country (str): country name
            scale (str): scale of geographic shapes, '10m', '50m' or '110m'

        Returns:
            geopandas.GeoDataFrame:
                Index
                    reset index
                Columns
                    - Value (int or float or None)
                    - geometry (geopandas.GeoDataFrame.geometry): geometry information
        """
        # Get geometry information of the country
        iso3 = self._to_iso3(country)
        scale = "50m" if iso3 == "USA" else "10m"
        gdf = self._load_geo_country_specific(scale=scale)
        gdf = gdf.loc[gdf[self.ISO3] == iso3]
        # Merge the data with geometry information
        gdf = gdf.merge(data, how="left", on=self.PROVINCE)
        # Select countries
        sel = set(included or gdf[self.PROVINCE].unique()) - set(excluded or [])
        return gdf.loc[gdf[self.PROVINCE].isin(sel), ["Value", "geometry"]]

    def _load_geo_global(self):
        """
        Retrieve geometry information for global map.

        Returns:
            geopandas.GeoDataFrame:
                Index
                    reset index
                Columns
                    - ISO3 (str): ISO3 codes
                    - geometry (geopandas.GeoDataFrame.geometry): geometry information
        """
        # Geometry information from Natural Earth
        # pop_est, continent, name, iso_a3, gdp_md_est, geometry
        geopath = gpd.datasets.get_path("naturalearth_lowres")
        gdf = gpd.read_file(geopath)
        # Data cleaning
        gdf["name"] = gdf["name"].apply(unidecode)
        gdf["name"].replace(
            {
                "Fr. S. Antarctic Lands": "French Southern and Antarctic Lands",
                "S. Sudan": "South Sudan"
            }, inplace=True)
        # Get ISO3 codes
        gdf.rename(columns={"iso_a3": self.ISO3}, inplace=True)
        sel = gdf[self.ISO3] == "-99"
        gdf.loc[sel, self.ISO3] = gdf.loc[sel, "name"].apply(self._to_iso3)
        # Remove Antarctica and
        gdf = gdf.loc[gdf[self.ISO3] != "ATA"]
        return gdf.loc[:, [self.ISO3, "geometry"]]

    def _load_geo_country_specific(self, scale):
        """
        Load shape file from 'Natural Earth Vector'.
        https://github.com/nvkelso/natural-earth-vector

        Args:
            scale (str): scale of geographic shapes, '10m', '50m' or '110m'

        Returns:
            geopandas.GeoDataFrame:
                Index
                    reset index
                Columns
                    - ISO3 (str): ISO3 codes
                    - Province (str): province name
                    - geometry (geopandas.GeoDataFrame.geometry): geometry information
        """
        title = f"ne_{scale}_admin_1_states_provinces"
        extensions = [".README.html", ".VERSION.txt", ".cpg", ".dbf", ".sbn", ".sbx", ".shp", ".shx"]
        geo_dirpath = self._geo_dirpath.joinpath(title)
        geo_dirpath.mkdir(parents=True, exist_ok=True)
        # Download files, if necessary
        for ext in extensions:
            basename = f"{title}{ext}"
            if geo_dirpath.joinpath(basename).exists():
                continue
            url = f"https://github.com/nvkelso/natural-earth-vector/blob/master/{scale}_cultural/{basename}?raw=true"
            response = requests.get(url=url)
            with geo_dirpath.joinpath(basename).open("wb") as fh:
                fh.write(response.content)
        # Data cleaning
        gdf = gpd.read_file(geo_dirpath.joinpath(f"{title}.shp"))
        gdf["name"] = gdf["name"].fillna("").apply(unidecode)
        gdf.rename(columns={"name": self.PROVINCE, "adm0_a3": self.ISO3}, inplace=True)
        return gdf.loc[:, [self.ISO3, self.PROVINCE, "geometry"]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.ode.mbase import ModelBase


class SEWIRF(ModelBase):
    """
    SEWIR-F model.
    """
    # Model name
    NAME = "SEWIR-F"
    # names of parameters
    PARAMETERS = ["theta", "kappa", "rho1", "rho2", "rho3", "sigma"]
    DAY_PARAMETERS = [
        "alpha1 [-]", "1/alpha2 [day]",
        "1/beta1 [day]", "1/beta2 [day]", "1/beta3 [day]",
        "1/gamma [day]"
    ]
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = {
        "x1": ModelBase.S,
        "y": ModelBase.CI,
        "z": ModelBase.R,
        "w": ModelBase.F,
        "x2": ModelBase.E,
        "x3": ModelBase.W,
    }
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array([0, 10, 10, 2, 0, 0])
    # Variables that increases monotonically
    VARS_INCLEASE = [ModelBase.R, ModelBase.F]
    # Example set of parameters and initial values
    EXAMPLE = {
        ModelBase.STEP_N: 180,
        ModelBase.N.lower(): 1_000_000,
        ModelBase.PARAM_DICT: {
            "theta": 0.002, "kappa": 0.005, "rho1": 0.2, "sigma": 0.075,
            "rho2": 0.167, "rho3": 0.167,
        },
        ModelBase.Y0_DICT: {
            ModelBase.S: 994_000, ModelBase.E: 3000, ModelBase.W: 0.002,
            ModelBase.CI: 1000, ModelBase.R: 0, ModelBase.F: 0,
        },
    }

    def __init__(self, population, theta, kappa, rho1, rho2, rho3, sigma):
        """

        Args:
        @population (int): total population
        parameter values of non-dimensional ODE model
            - @theta (float)
            - @kappa (float)
            - @rho1 (float)
            - @rho2 (float)
            - @rho3 (float)
            - @sigma (float)
        """
        # Total population
        self.population = self._ensure_natural_int(
            population, name="population"
        )
        # Non-dim parameters
        self.theta = theta
        self.kappa = kappa
        self.rho1 = rho1
        self.rho2 = rho2
        self.rho3 = rho3
        self.sigma = sigma
        self.non_param_dict = {
            "theta": theta, "kappa": kappa,
            "rho1": rho1, "rho2": rho2, "rho3": rho3,
            "sigma": sigma
        }

    def __call__(self, t, X):
        """
        Return the list of dS/dt (tau-free) etc.

        Args:
            t (int): time steps
            X (numpy.array): values of th model variables

        Returns:
            (np.array)
        """
        n = self.population
        s, i, *_, e, w = X
        beta_swi = self.rho1 * s * (w + i) / n
        dsdt = 0 - beta_swi
        dedt = beta_swi - self.rho2 * e
        dwdt = self.rho2 * e - self.rho3 * w
        drdt = self.sigma * i
        dfdt = self.kappa * i + self.theta * self.rho3 * w
        didt = 0 - dsdt - drdt - dfdt - dedt - dwdt
        return np.array([dsdt, didt, drdt, dfdt, dedt, dwdt])

    @classmethod
    def param_range(cls, taufree_df, population):
        """
        Define the range of parameters (not including tau value).

        Args:
            taufree_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): time steps (tau-free)
                    - columns with dimensional variables
            population (int): total population

        Returns:
            (dict)
                - key (str): parameter name
                - value (tuple(float, float)): min value and max value
        """
        raise NotImplementedError(
            "SEWIR-F cannot be used for parameter estimation because we do not have records "
            "of Exposed and Waiting. Please use SIR-F model with `covsirphy.SIRF` class."
        )

    @classmethod
    def specialize(cls, data_df, population):
        """
        Specialize the dataset for this model.

        Args:
            data_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any columns
            population (int): total population in the place

        Returns:
            (pandas.DataFrame)
                Index
                    reset index
                Columns
                    - any columns @data_df has
                    - Susceptible (int): the number of susceptible cases
                    - Exposed (int): 0
                    - Waiting (int): 0
        """
        df = cls._ensure_dataframe(
            data_df, name="data_df", columns=cls.VALUE_COLUMNS)
        # Calculate dimensional variables
        df[cls.S] = population - df[cls.C]
        df[cls.E] = 0
        df[cls.W] = 0
        return df

    @classmethod
    def restore(cls, specialized_df):
        """
        Restore Confirmed/Infected/Recovered/Fatal.
         using a dataframe with the variables of the model.

        Args:
            specialized_df (pandas.DataFrame): dataframe with the variables

                Index
                    (object)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
                    - Exposed (int): Exposed and in latent period (without infectivity)
                    - Waiting (int): Waiting cases for confirmation (with infectivity)
                    - any columns

        Returns:
            (pandas.DataFrame):
                Index
                    (object): as-is
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - the other columns @specialzed_df has
        """
        df = specialized_df.copy()
        other_cols = list(set(df.columns) - set(cls.VALUE_COLUMNS))
        df[cls.C] = df[cls.CI] + df[cls.R] + df[cls.F]
        return df.loc[:, [*cls.VALUE_COLUMNS, *other_cols]]

    def calc_r0(self):
        """
        Calculate (basic) reproduction number.

        Returns:
            float
        """
        try:
            rho = self.rho1 / self.rho2 * self.rho3
            rt = rho * (1 - self.theta) / (self.sigma + self.kappa)
        except ZeroDivisionError:
            return None
        return round(rt, 2)

    def calc_days_dict(self, tau):
        """
        Calculate 1/beta [day] etc.

        Args:
            param tau (int): tau value [min]

        Returns:
            dict[str, int]
        """
        try:
            return {
                "alpha1 [-]": round(self.theta, 3),
                "1/alpha2 [day]": int(tau / 24 / 60 / self.kappa),
                "1/beta1 [day]": int(tau / 24 / 60 / self.rho1),
                "1/beta2 [day]": int(tau / 24 / 60 / self.rho2),
                "1/beta3 [day]": int(tau / 24 / 60 / self.rho3),
                "1/gamma [day]": int(tau / 24 / 60 / self.sigma)
            }
        except ZeroDivisionError:
            return {p: None for p in self.DAY_PARAMETERS}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import numpy as np
import pandas as pd
from covsirphy.cleaning.country_data import CountryData


class JapanData(CountryData):
    """
    Japan-specific dataset.

    Args:
        filename (str or pathlib.path): CSV filename to save the raw dataset
        force (bool): if True, always download the dataset from the server
        verbose (int): level of verbosity

    Note:
        Columns of JapanData.cleaned():
            - Date (pandas.TimeStamp): date
            - Country (pandas.Category): 'Japan'
            - Province (pandas.Category): '-' (country level), 'Entering' or province names
            - Confirmed (int): the number of confirmed cases
            - Infected (int): the number of currently infected cases
            - Fatal (int): the number of fatal cases
            - Recovered (int): the number of recovered cases
            - Tests (int): the number of tested persons
            - Moderate (int): the number of cases who requires hospitalization but not severe
            - Severe (int): the number of severe cases
            - Vaccinations (int): cumulative number of vaccinations
            - Vaccinated_once (int): cumulative number of people who received at least one vaccine dose
            - Vaccinated_full (int): cumulative number of people who received all doses prescrived by the protocol
    """
    GITHUB_URL = "https://raw.githubusercontent.com"
    URL_C = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_total.csv"
    URL_P = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_prefecture.csv"
    URL_M = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_metadata.csv"
    # Moderate: cases who requires hospitalization but not severe
    MODERATE = "Moderate"
    # Severe
    SEVERE = "Severe"
    # Column names
    JAPAN_VALUE_COLS = [
        CountryData.C, CountryData.CI, CountryData.F, CountryData.R,
        CountryData.TESTS, MODERATE, SEVERE,
        CountryData.VAC, CountryData.V_ONCE, CountryData.V_FULL,
    ]
    JAPAN_COLS = [
        CountryData.DATE, CountryData.COUNTRY, CountryData.PROVINCE,
        *JAPAN_VALUE_COLS,
    ]
    # Meta data
    JAPAN_META_CAT = ["Prefecture", "Admin_Capital", "Admin_Region"]
    JAPAN_META_INT = [
        "Admin_Num", "Area_Habitable", "Area_Total",
        "Clinic_bed_Care", "Clinic_bed_Total",
        "Hospital_bed_Care", "Hospital_bed_Specific", "Hospital_bed_Total",
        "Hospital_bed_Tuberculosis", "Hospital_bed_Type-I", "Hospital_bed_Type-II",
        "Population_Female", "Population_Male", "Population_Total",
    ]
    JAPAN_META_FLT = ["Location_Latitude", "Location_Longitude"]
    JAPAN_META_COLS = [*JAPAN_META_CAT, *JAPAN_META_INT, *JAPAN_META_FLT]

    def __init__(self, filename, force=False, verbose=1):
        Path(filename).parent.mkdir(exist_ok=True, parents=True)
        if Path(filename).exists() and not force:
            try:
                self._raw = self.load(filename)
            except KeyError:
                self._raw = self._retrieve(filename=filename, verbose=verbose)
        else:
            self._raw = self._retrieve(filename=filename, verbose=verbose)
        self._cleaned_df = self._cleaning()
        self._country = "Japan"
        self._citation = "Lisphilar (2020), COVID-19 dataset in Japan, GitHub repository, " \
            "https://github.com/lisphilar/covid19-sir/data/japan"
        self.dir_path = Path(filename).parent
        self.verbose = verbose
        # Directory that save the file
        self._dirpath = Path(filename or "input").resolve().parent

    def _retrieve(self, filename, verbose=1):
        """
        Retrieve the dataset from server.

        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            verbose (int): level of verbosity

        Returns:
            pd.DataFrame: raw dataset
        """
        # Show URL
        if verbose:
            print("Retrieving COVID-19 dataset in Japan from https://github.com/lisphilar/covid19-sir/data/japan")
        # Download the dataset at country level
        cols = [
            "Area", "Date", "Positive",
            "Tested", "Discharged", "Fatal", "Hosp_require", "Hosp_severe",
        ]
        cols_v = ["Vaccinated_1st", "Vaccinated_2nd"]
        c_df = self.load(self.URL_C, header=0).rename({"Location": "Area"}, axis=1)[cols + cols_v]
        # Download the datset at province level
        p_df = self.load(self.URL_P, header=0).rename({"Prefecture": "Area"}, axis=1)[cols]
        # Combine the datsets
        df = pd.concat([c_df, p_df], axis=0, ignore_index=True, sort=True)
        # Save the raw data
        df.to_csv(filename, index=False)
        return df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame: cleaned data
        """
        df = self._raw.copy()
        # Rename columns
        df = df.rename(
            {
                "Area": self.PROVINCE,
                "Date": self.DATE,
                "Positive": self.C,
                "Fatal": self.F,
                "Discharged": self.R,
                "Hosp_severe": self.SEVERE,
                "Tested": self.TESTS
            },
            axis=1
        )
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Fill NA values
        for col in [self.C, self.F, self.R, self.SEVERE, "Hosp_require", self.TESTS]:
            df[col] = pd.to_numeric(df[col], errors="coerce")
        df = df.groupby(self.PROVINCE).apply(
            lambda x: x.set_index(self.DATE).resample("D").interpolate("linear", limit_direction="both"))
        df = df.fillna(0).drop(self.PROVINCE, axis=1).reset_index()
        df = df.sort_values(self.DATE).reset_index(drop=True)
        # Moderate
        df[self.MODERATE] = df["Hosp_require"] - df[self.SEVERE]
        # Vaccinations
        v_raw_cols = ["Vaccinated_1st", "Vaccinated_2nd"]
        v_df = self._raw.loc[:, ["Area", "Date", *v_raw_cols]].rename(
            columns={"Area": self.PROVINCE, "Date": self.DATE})
        v_df[self.DATE] = pd.to_datetime(v_df[self.DATE])
        for col in v_raw_cols:
            v_df[col] = pd.to_numeric(v_df[col], errors="coerce").fillna(0)
        v_1st = v_df.groupby(self.PROVINCE)["Vaccinated_1st"].cumsum().fillna(0)
        v_2nd = v_df.groupby(self.PROVINCE)["Vaccinated_2nd"].cumsum().fillna(0)
        v_sum_df = pd.DataFrame(
            {
                self.PROVINCE: v_df[self.PROVINCE],
                self.DATE: v_df[self.DATE],
                self.VAC: v_1st + v_2nd,
                self.V_ONCE: v_1st,
                self.V_FULL: v_2nd,
            }
        )
        df = df.drop(v_raw_cols, axis=1).merge(
            v_sum_df, how="left", on=[self.PROVINCE, self.DATE]).reset_index(drop=True)
        # Records at country level (Domestic/Airport/Returnee) and entering Japan (Airport/Returnee)
        e_cols = ["Airport", "Returnee"]
        e_df = df.loc[df[self.PROVINCE].isin(e_cols)].groupby(self.DATE).sum()
        e_df[self.PROVINCE] = "Entering"
        c_cols = ["Domestic", "Airport", "Returnee"]
        c_df = df.loc[df[self.PROVINCE].isin(c_cols)].groupby(self.DATE).sum()
        c_df[self.PROVINCE] = self.UNKNOWN
        df = pd.concat(
            [
                df.loc[~df[self.PROVINCE].isin(c_cols)],
                e_df.reset_index(),
                c_df.reset_index(),
            ],
            ignore_index=True, sort=True)
        # Value columns
        df[self.CI] = df[self.C] - df[self.F] - df[self.R]
        df[self.JAPAN_VALUE_COLS] = df[self.JAPAN_VALUE_COLS].astype(np.int64)
        # Country
        df[self.COUNTRY] = "Japan"
        # Update data types to reduce memory
        df[self.AREA_COLUMNS] = df[self.AREA_COLUMNS].astype("category")
        return df.loc[:, self.JAPAN_COLS]

    def set_variables(self):
        raise NotImplementedError

    def _retrieve_meta(self, filename, verbose=1):
        """
        Retrieve meta data from server.

        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame: raw dataset
        """
        # Show URL
        if verbose:
            print("Retrieving Metadata of Japan dataset from https://github.com/lisphilar/covid19-sir/data/japan")
        df = self.load(self.URL_M)
        df.to_csv(filename, index=False)
        return df

    def meta(self, basename="covid_japan_metadata.csv", cleaned=True, force=False):
        """
        Return metadata of Japan-specific dataset.

        Args:
            basename (str): basename of the CSV file to save the raw dataset
            cleaned (bool): return cleaned (True) or raw (False) dataset
            force (bool): if True, always download the dataset from the server

        Returns:
            pandas.DataFrame: (cleaned or raw) dataset
                Index
                    reset index
                Columns for cleaned dataset,
                    - Prefecture (pandas.Category)
                    - Admin_Capital (pandas.Category)
                    - Admin_Region (pandas.Category)
                    - Admin_Num (int)
                    - Area_Habitable (int)
                    - Area_Total (int)
                    - Clinic_bed_Care (int)
                    - Clinic_bed_Total (int)
                    - Hospital_bed_Care (int)
                    - Hospital_bed_Specific (int)
                    - Hospital_bed_Total (int)
                    - Hospital_bed_Tuberculosis (int)
                    - Hospital_bed_Type-I (int)
                    - Hospital_bed_Type-II (int)
                    - Population_Female (int)
                    - Population_Male (int)
                    - Population_Total (int)
                    - Location_Latitude (float)
                    - Location_Longitude (float)

        Note:
            Please refer to https://github.com/lisphilar/covid19-sir/tree/master/data
        """
        filepath = self.dir_path.joinpath(basename)
        if filepath.exists() and not force:
            df = self.load(filepath)
        else:
            df = self._retrieve_meta(filename=filepath, verbose=self.verbose)
        if not cleaned:
            return df
        # Data cleaning
        df["Title"] = df["Category"].str.cat(df["Item"], sep="_")
        df = df.pivot_table(
            values="Value", index="Prefecture", columns="Title", aggfunc="last")
        df.columns.name = None
        df = df.reset_index()
        df[self.JAPAN_META_CAT] = df[self.JAPAN_META_CAT].astype("category")
        df[self.JAPAN_META_INT] = df[self.JAPAN_META_INT].astype(np.int64)
        df[self.JAPAN_META_FLT] = df[self.JAPAN_META_FLT].astype(np.float64)
        df = df.sort_values("Admin_Num").reset_index(drop=True)
        return df.loc[:, self.JAPAN_META_COLS]
